# Generated by Django 2.2.5 on 2019-11-25 21:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structures', '0002_auto_20191121_2335'),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='is_timer_added',
            field=models.BooleanField(blank=True, default=None, help_text='True when a timer has been added for this notification', null=True),
        ),
        migrations.AlterField(
            model_name='notification',
            name='is_sent',
            field=models.BooleanField(blank=True, default=False, help_text='True when this notification has been forwarded to Discord'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='last_updated',
            field=models.DateTimeField(help_text='Date when this notification has last been updated from ESI'),
        ),
    ]



# Generated by Django 2.2.8 on 2020-01-20 01:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structures', '0005_notification_created'),
    ]

    operations = [
        migrations.CreateModel(
            name='StructureTag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='name of the tag - must be unique', max_length=255, unique=True)),
                ('description', models.TextField(blank=True, default=None, help_text='description for this tag', null=True)),
                ('style', models.CharField(blank=True, choices=[('default', 'grey'), ('primary', 'dark blue'), ('success', 'green'), ('info', 'light blue'), ('warning', 'yellow'), ('danger', 'red')], default='default', help_text='color style of tag', max_length=16)),
                ('is_default', models.BooleanField(default=False, help_text='if true this tag will automatically added to new structures')),
            ],
        ),
        migrations.AddField(
            model_name='structure',
            name='tags',
            field=models.ManyToManyField(blank=True, default=None, help_text='list of tags for this structure', to='structures.StructureTag'),
        ),
    ]



import logging
from pydoc import locate

from bravado.exception import HTTPError

from django.db import models, transaction
from django.utils.timezone import now

from esi.models import Token

from . import __title__
from structures.helpers.esi_fetch import esi_fetch_with_localization, esi_fetch
from .utils import LoggerAddTag, make_logger_prefix


logger = LoggerAddTag(logging.getLogger(__name__), __title__)


class EveUniverseManager(models.Manager):
    
    def get_or_create_esi(self, eve_id: int) -> tuple:
        """gets or creates eve universe object fetched from ESI if needed. 
        Will always get/create parent objects.
        
        eve_id: Eve Online ID of object

        Returns: object, created        
        """
        try:
            obj = self.get(id=eve_id)
            created = False        
        except self.model.DoesNotExist:
            obj, created = self.update_or_create_esi(eve_id)

        return obj, created

    def update_or_create_esi(self, eve_id: int) -> tuple:
        """updates or creates Eve Universe object with data fetched from ESI. 
        Will always update/create children and get/create parent objects.

        eve_id: Eve Online ID of object

        Returns: object, created
        """
        from .models import EsiNameLocalization
        add_prefix = make_logger_prefix(
            '%s(id=%d)' % (self.model.__name__, eve_id)
        )        
        try:            
            esi_path = 'Universe.' + self.model.esi_method()
            args = {self.model.esi_pk(): eve_id}
            if self.model.has_esi_localization():
                eve_data_objects = \
                    esi_fetch_with_localization(
                        esi_path=esi_path,
                        languages=EsiNameLocalization.ESI_LANGUAGES,
                        args=args,                        
                        logger_tag=add_prefix()
                    )
            else:
                eve_data_objects = dict()
                eve_data_objects[EsiNameLocalization.ESI_DEFAULT_LANGUAGE] = \
                    esi_fetch(
                        esi_path=esi_path,
                        args=args,
                        logger_tag=add_prefix()
                    )   # noqa E123
            defaults = self.model.map_esi_fields_to_model(eve_data_objects)
            obj, created = self.update_or_create(id=eve_id, defaults=defaults)
            obj.set_generated_translations()
            obj.save()
            self._update_or_create_children(eve_data_objects)
        
        except Exception as ex:
            logger.warn(add_prefix('Failed to update or create: %s' % ex))
            raise ex

        return obj, created
        
    def _update_or_create_children(self, eve_data_objects: dict) -> None:
        """updates or creates child objects if specified"""
        eve_data_obj = eve_data_objects[self.model.ESI_DEFAULT_LANGUAGE]
        for key, child_class in self.model.child_mappings().items():
            ChildClass = locate(__package__ + '.models.' + child_class)
            for eve_data_obj_2 in eve_data_obj[key]:
                eve_id = eve_data_obj_2[ChildClass.esi_pk()]
                ChildClass.objects.update_or_create_esi(eve_id)                
                
    def update_all_esi(self) -> int:
        """update all objects from ESI. Returns count of updated  objects"""
        logger.info(
            '%s: Updating %d objects from from ESI...' 
            % (self.model.__name__, self.count())
        )
        count_updated = 0
        for eve_obj in self.all().order_by('last_updated'):
            try:
                self.update_or_create_esi(eve_obj.id)
                count_updated += 1
            except HTTPError as ex:
                logger.exception('Update interrupted by exception: %s' % ex)

        return count_updated


class EveSovereigntyMapManager(models.Manager):
    
    def update_from_esi(self):        
        logger.info('Fetching sovereignty map from ESI...')
        map = esi_fetch('Sovereignty.get_sovereignty_map', args={})
        last_updated = now()
        obj_list = list()
        for solar_system in map:
            obj_def = {
                'solar_system_id': solar_system['system_id'],
                'last_updated': last_updated
            }
            for key in ['alliance_id', 'corporation_id', 'faction_id']:
                if key in solar_system and solar_system[key]:
                    obj_def[key] = solar_system[key]
                else:
                    obj_def[key] = None

            if (
                obj_def['alliance_id']
                or obj_def['corporation_id']
                or obj_def['faction_id']
            ):                
                obj_list.append(self.model(**obj_def))
        
        if obj_list:
            logger.info('Storing sovereignty map ...')
            with transaction.atomic():
                self.all().delete()
                self.bulk_create(obj_list, batch_size=1000)
                    

class EveEntityManager(models.Manager):

    def get_or_create_esi(self, eve_entity_id: int) -> tuple:
        """gets or creates EveEntity obj with data fetched from ESI if needed
        
        eve_id: Eve Online ID of object
        
        Returns: object, created        
        """        
        try:
            obj = self.get(id=eve_entity_id)
            created = False
        except self.model.DoesNotExist:
            obj, created = self.update_or_create_esi(eve_entity_id)

        return obj, created

    def update_or_create_esi(self, eve_entity_id: int) -> tuple:
        """updates or creates EveEntity object with data fetched from ESI
        
        eve_id: Eve Online ID of object
        
        Returns: object, created        
        """        
        add_prefix = make_logger_prefix(
            '%s(id=%d)' % (self.model.__name__, eve_entity_id)
        )        
        try:            
            response = esi_fetch(
                esi_path='Universe.post_universe_names', 
                args={'ids': [eve_entity_id]},
                logger_tag=add_prefix()
            )
            if len(response) > 0:
                first = response[0]
                category = self.model.get_matching_entity_category(
                    first['category']
                )
                obj, created = self.update_or_create(
                    id=eve_entity_id,
                    defaults={
                        'category': category,
                        'name': first['name']
                    }
                )
            else:
                raise ValueError(add_prefix('Did not find a match'))

        except Exception as ex:
            logger.warn(add_prefix('Failed to load eve entity: %s' % ex))
            raise ex

        return obj, created


class StructureManager(models.Manager):

    def get_or_create_esi(
        self, structure_id: int, token: Token
    ) -> tuple:
        """get or create a structure with data from ESI if needed
        
        structure_id: Structure ID of object in Eve Online

        token: ``esi.models.Token`` object with scope: 
        ``esi-universe.read_structures.v1``
        
        Returns: object, created
        """        
        try:
            obj = self.get(id=structure_id)
            created = False
        except self.model.DoesNotExist:            
            obj, created = self.update_or_create_esi(structure_id, token)
        return obj, created

    def update_or_create_esi(
        self, structure_id: int, token: Token
    ) -> tuple:
        """update or create a structure from ESI for given structure ID
        This will only fetch basic info about a structure
        
        structure_id: Structure ID of object in Eve Online

        token: ``esi.models.Token`` object with scope: 
        ``esi-universe.read_structures.v1``
        
        Returns: object, created
        """
        from .models import Owner

        add_prefix = make_logger_prefix(
            '%s(id=%d)' % (self.model.__name__, structure_id)
        )
        logger.info(add_prefix('Trying to fetch structure from ESI'))
        
        try:
            if token is None:
                raise ValueError('Can not fetch structure without token')
            
            structure_info = esi_fetch(
                esi_path='Universe.get_universe_structures_structure_id', 
                args={'structure_id': structure_id},                 
                token=token,
                logger_tag=add_prefix()
            )            
            structure = {
                'structure_id': structure_id,
                'name': self.model.extract_name_from_esi_respose(
                    structure_info['name']
                ),
                'position': structure_info['position'],
                'type_id': structure_info['type_id'],
                'system_id': structure_info['solar_system_id']
            }
            owner = Owner.objects.get(
                corporation__corporation_id=structure_info['corporation_id']
            )
            obj, created = self.update_or_create_from_dict(
                structure=structure, owner=owner
            )
        
        except Exception as ex:
            logger.warn(add_prefix('Failed to load structure: {}'.format(ex)))
            raise ex

        return obj, created

    def update_or_create_from_dict(
        self, structure: dict, owner: object
    ) -> tuple:
        """update or create structure from given dict"""
        from .models import (
            EveType, 
            EveSolarSystem, 
            StructureService, 
            EvePlanet, 
            EveMoon
        )
        from .models.eveuniverse import EveUniverse
        eve_type, _ = EveType.objects.get_or_create_esi(structure['type_id'])
        eve_solar_system, _ = \
            EveSolarSystem.objects.get_or_create_esi(structure['system_id'])
        fuel_expires = \
            structure['fuel_expires'] \
            if 'fuel_expires' in structure else None

        next_reinforce_hour = \
            structure['next_reinforce_hour']  \
            if 'next_reinforce_hour' in structure else None

        next_reinforce_weekday = \
            structure['next_reinforce_weekday'] \
            if 'next_reinforce_weekday' in structure else None

        next_reinforce_apply = \
            structure['next_reinforce_apply'] \
            if 'next_reinforce_apply' in structure else None

        reinforce_hour = \
            structure['reinforce_hour'] \
            if 'reinforce_hour' in structure else None

        state = \
            self.model.get_matching_state_for_esi_state(structure['state']) \
            if 'state' in structure else self.model.STATE_UNKNOWN

        state_timer_start = \
            structure['state_timer_start'] \
            if 'state_timer_start' in structure else None

        state_timer_end = \
            structure['state_timer_end'] \
            if 'state_timer_end' in structure else None

        unanchors_at =  \
            structure['unanchors_at']\
            if 'unanchors_at' in structure else None

        position_x = \
            structure['position']['x']\
            if 'position' in structure else None

        position_y = \
            structure['position']['y']\
            if 'position' in structure else None

        position_z = \
            structure['position']['z']\
            if 'position' in structure else None

        if 'planet_id' in structure:
            eve_planet, _ = \
                EvePlanet.objects.get_or_create_esi(structure['planet_id'])
        else:
            eve_planet = None

        if 'moon_id' in structure:
            eve_moon, _ = \
                EveMoon.objects.get_or_create_esi(structure['moon_id'])
        else:
            eve_moon = None

        obj, created = self.update_or_create(
            id=structure['structure_id'],
            defaults={
                'owner': owner,
                'eve_type': eve_type,
                'name': structure['name'],
                'eve_solar_system': eve_solar_system,
                'eve_planet': eve_planet,
                'eve_moon': eve_moon,
                'position_x': position_x,
                'position_y': position_y,
                'position_z': position_z,
                'fuel_expires': fuel_expires,
                'next_reinforce_hour': next_reinforce_hour,
                'next_reinforce_weekday': next_reinforce_weekday,
                'next_reinforce_apply': next_reinforce_apply,
                'reinforce_hour': reinforce_hour,
                'state': state,
                'state_timer_start': state_timer_start,
                'state_timer_end': state_timer_end,
                'unanchors_at': unanchors_at,
                'last_updated': owner.structures_last_sync
            }
        )
        # save related structure services
        StructureService.objects.filter(structure=obj).delete()
        if 'services' in structure and structure['services']:
            for service in structure['services']:
                state = StructureService.get_matching_state_for_esi_state(
                    service['state']
                )
                args = {
                    'structure': obj,
                    'name': service['name'],
                    'state': state
                }
                for lang in EveUniverse.ESI_LANGUAGES:
                    if lang != EveUniverse.ESI_DEFAULT_LANGUAGE:
                        field_name = 'name_%s' % lang
                        if field_name in service:
                            args[field_name] = service[field_name]

                StructureService.objects.create(**args)
                
        return obj, created


class StructureTagManager(models.Manager):
    
    def get_or_create_for_space_type(self, solar_system: object) -> tuple:
        if solar_system.space_type in self.model.SPACE_TYPE_MAP:
            name = self.model.SPACE_TYPE_MAP[solar_system.space_type]['name']
        else:
            name = None

        if name:
            try:
                obj = self.get(name=name)
                created = False
            except self.model.DoesNotExist:
                obj, created = self.update_or_create_for_space_type(solar_system)

        else:
            obj = None
            created = None
        
        return obj, created

    def update_or_create_for_space_type(self, solar_system: object) -> tuple:
        if solar_system.space_type in self.model.SPACE_TYPE_MAP:
            params = self.model.SPACE_TYPE_MAP[solar_system.space_type]            
        else:
            params = None
        
        if params:
            return self.update_or_create(
                name=params['name'], 
                defaults={
                    'style': params['style'],
                    'description': (
                        'this tag represents a space type. system generated.'
                    ),                
                    'order': 50,
                    'is_user_managed': False,
                    'is_default': False
                }
            )
        else:
            return None, None

    def get_or_create_for_sov(self) -> tuple:
        try:
            obj = self.get(name=self.model.NAME_SOV_TAG)
            created = False
        except self.model.DoesNotExist:
            obj, created = self.update_or_create_for_sov()

        return obj, created
    
    def update_or_create_for_sov(self) -> tuple:
        obj, created = self.update_or_create(
            name=self.model.NAME_SOV_TAG, 
            defaults={
                'style': self.model.STYLE_DARK_BLUE,
                'description': (
                    'Owner of this structure has sovereignty. system generated.'
                ),
                'order': 20,
                'is_user_managed': False,
                'is_default': False
            }
        )
        return obj, created



from django.apps import AppConfig
from . import __version__


class StructuresConfig(AppConfig):
    name = 'structures'
    label = 'structures'
    verbose_name = 'Structures v{}'.format(__version__)

    def ready(self):
        from . import signals  # noqa: F401



from django.utils.translation import ugettext_lazy as _

from allianceauth.services.hooks import MenuItemHook, UrlHook
from allianceauth import hooks

from . import urls, __title__


class StructuresMenuItem(MenuItemHook):
    """ This class ensures only authorized users will see the menu entry """
    def __init__(self):
        # setup menu entry for sidebar
        MenuItemHook.__init__(
            self,
            _(__title__),
            'fa fa-building fa-fw',
            'structures:index',
            navactive=['structures:index']
        )

    def render(self, request):
        if request.user.has_perm('structures.basic_access'):
            return MenuItemHook.render(self, request)
        return ''


@hooks.register('menu_item_hook')
def register_menu():
    return StructuresMenuItem()


@hooks.register('url_hook')
def register_urls():
    return UrlHook(urls, 'structures', r'^structures/')



"""Notification related models"""

from datetime import datetime, timedelta
import json
import logging
from time import sleep
import yaml

import pytz
import dhooks_lite

from django.db import models
from django.core.validators import MinValueValidator
from django.conf import settings
from django.utils.translation import gettext_lazy as _, gettext
from django.utils import translation

from allianceauth.eveonline.evelinks import dotlan

from esi.models import Token
from multiselectfield import MultiSelectField

from ..app_settings import (
    STRUCTURES_DEFAULT_LANGUAGE,
    STRUCTURES_DEVELOPER_MODE,
    STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED,
    STRUCTURES_NOTIFICATION_MAX_RETRIES,    
    STRUCTURES_NOTIFICATION_WAIT_SEC,    
    STRUCTURES_REPORT_NPC_ATTACKS,    
    STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
)
from .. import __title__
from ..managers import EveEntityManager
from ..utils import (
    LoggerAddTag, DATETIME_FORMAT, make_logger_prefix, app_labels
)
from .eveuniverse import EveType, EveSolarSystem, EveMoon, EvePlanet
from .structures import Structure

if 'timerboard' in app_labels():
    from allianceauth.timerboard.models import Timer

logger = LoggerAddTag(logging.getLogger(__name__), __title__)

# Supported languages
LANGUAGES = (
    ('en', _('English')),
    ('de', _('German')),
    ('es', _('Spanish')),
    ('zh-hans', _('Chinese Simplified')),
    ('ru', _('Russian')),
    ('ko', _('Korean')),
)

# Notification types
NTYPE_MOONS_AUTOMATIC_FRACTURE = 401
NTYPE_MOONS_EXTRACTION_CANCELED = 402
NTYPE_MOONS_EXTRACTION_FINISHED = 403
NTYPE_MOONS_EXTRACTION_STARTED = 404
NTYPE_MOONS_LASER_FIRED = 405

NTYPE_STRUCTURE_ANCHORING = 501
NTYPE_STRUCTURE_DESTROYED = 502
NTYPE_STRUCTURE_FUEL_ALERT = 503
NTYPE_STRUCTURE_LOST_ARMOR = 504
NTYPE_STRUCTURE_LOST_SHIELD = 505
NTYPE_STRUCTURE_ONLINE = 506
NTYPE_STRUCTURE_SERVICES_OFFLINE = 507
NTYPE_STRUCTURE_UNANCHORING = 508
NTYPE_STRUCTURE_UNDER_ATTACK = 509
NTYPE_STRUCTURE_WENT_HIGH_POWER = 510
NTYPE_STRUCTURE_WENT_LOW_POWER = 511
NTYPE_STRUCTURE_REINFORCE_CHANGED = 512
NTYPE_OWNERSHIP_TRANSFERRED = 513

NTYPE_ORBITAL_ATTACKED = 601
NTYPE_ORBITAL_REINFORCED = 602

NTYPE_TOWER_ALERT_MSG = 701
NTYPE_TOWER_RESOURCE_ALERT_MSG = 702

NTYPE_SOV_ENTOSIS_CAPTURE_STARTED = 801
NTYPE_SOV_COMMAND_NODE_EVENT_STARTED = 802
NTYPE_SOV_ALL_CLAIM_ACQUIRED_MSG = 803
NTYPE_SOV_STRUCTURE_REINFORCED = 804
NTYPE_SOV_STRUCTURE_DESTROYED = 805

NTYPE_CHOICES = [
    # moon mining
    (NTYPE_MOONS_AUTOMATIC_FRACTURE, 'MoonminingAutomaticFracture'),
    (NTYPE_MOONS_EXTRACTION_CANCELED, 'MoonminingExtractionCancelled'),
    (NTYPE_MOONS_EXTRACTION_FINISHED, 'MoonminingExtractionFinished'),
    (NTYPE_MOONS_EXTRACTION_STARTED, 'MoonminingExtractionStarted'),
    (NTYPE_MOONS_LASER_FIRED, 'MoonminingLaserFired'),

    # upwell structures general
    (NTYPE_OWNERSHIP_TRANSFERRED, 'OwnershipTransferred'),
    (NTYPE_STRUCTURE_ANCHORING, 'StructureAnchoring'),
    (NTYPE_STRUCTURE_DESTROYED, 'StructureDestroyed'),
    (NTYPE_STRUCTURE_FUEL_ALERT, 'StructureFuelAlert'),
    (NTYPE_STRUCTURE_LOST_ARMOR, 'StructureLostArmor'),
    (NTYPE_STRUCTURE_LOST_SHIELD, 'StructureLostShields'),
    (NTYPE_STRUCTURE_ONLINE, 'StructureOnline'),
    (NTYPE_STRUCTURE_SERVICES_OFFLINE, 'StructureServicesOffline'),
    (NTYPE_STRUCTURE_UNANCHORING, 'StructureUnanchoring'),
    (NTYPE_STRUCTURE_UNDER_ATTACK, 'StructureUnderAttack'),
    (NTYPE_STRUCTURE_WENT_HIGH_POWER, 'StructureWentHighPower'),
    (NTYPE_STRUCTURE_WENT_LOW_POWER, 'StructureWentLowPower'),

    # custom offices only
    (NTYPE_ORBITAL_ATTACKED, 'OrbitalAttacked'),
    (NTYPE_ORBITAL_REINFORCED, 'OrbitalReinforced'),

    # starbases only
    (NTYPE_TOWER_ALERT_MSG, 'TowerAlertMsg'),
    (NTYPE_TOWER_RESOURCE_ALERT_MSG, 'TowerResourceAlertMsg'),

    # sov
    (NTYPE_SOV_ENTOSIS_CAPTURE_STARTED, 'EntosisCaptureStarted'),
    (NTYPE_SOV_COMMAND_NODE_EVENT_STARTED, 'SovCommandNodeEventStarted'),
    (NTYPE_SOV_ALL_CLAIM_ACQUIRED_MSG, 'SovAllClaimAquiredMsg'),
    (NTYPE_SOV_STRUCTURE_REINFORCED, 'SovStructureReinforced'),
    (NTYPE_SOV_STRUCTURE_DESTROYED, 'SovStructureDestroyed'),
]

_NTYPE_RELEVANT_FOR_TIMERBOARD = [
    NTYPE_STRUCTURE_LOST_SHIELD,
    NTYPE_STRUCTURE_LOST_ARMOR,
    NTYPE_STRUCTURE_ANCHORING,
    NTYPE_ORBITAL_REINFORCED,
    NTYPE_MOONS_EXTRACTION_STARTED,
    NTYPE_MOONS_EXTRACTION_CANCELED,
    NTYPE_SOV_STRUCTURE_REINFORCED
]

NTYPE_FOR_ALLIANCE_LEVEL = [
    NTYPE_SOV_ENTOSIS_CAPTURE_STARTED,
    NTYPE_SOV_COMMAND_NODE_EVENT_STARTED,
    NTYPE_SOV_ALL_CLAIM_ACQUIRED_MSG,
    NTYPE_SOV_STRUCTURE_REINFORCED,
    NTYPE_SOV_STRUCTURE_DESTROYED
]


def get_default_notification_types():
    """generates a set of all existing notification types as default"""
    return tuple(sorted([str(x[0]) for x in NTYPE_CHOICES]))


class Webhook(models.Model):
    """A destination for forwarding notification alerts"""

    TYPE_DISCORD = 1

    TYPE_CHOICES = [
        (TYPE_DISCORD, _('Discord Webhook')),
    ]

    name = models.CharField(
        max_length=64,
        unique=True,
        help_text='short name to identify this webhook'
    )
    webhook_type = models.IntegerField(
        choices=TYPE_CHOICES,
        default=TYPE_DISCORD,
        help_text='type of this webhook'
    )
    url = models.CharField(
        max_length=255,
        unique=True,
        help_text=(
            'URL of this webhook, e.g. '
            'https://discordapp.com/api/webhooks/123456/abcdef'
        )
    )
    notes = models.TextField(
        null=True,
        default=None,
        blank=True,
        help_text='you can add notes about this webhook here if you want'
    )
    notification_types = MultiSelectField(
        choices=NTYPE_CHOICES,
        default=get_default_notification_types,
        help_text=(
            'only notifications which selected types are sent to this webhook'
        )
    )
    language_code = models.CharField(
        max_length=8, 
        choices=LANGUAGES,
        default=None,
        null=True,
        blank=True,
        verbose_name='language',
        help_text='language of notifications send to this webhook'
    )
    is_active = models.BooleanField(
        default=True,
        help_text='whether notifications are currently sent to this webhook'
    )
    is_default = models.BooleanField(
        default=False,
        help_text=(
            'whether owners have this webhook automatically '
            'pre-set when created'
        )
    )

    def __str__(self):
        return self.name

    def __repr__(self):
        return '{}(id={}, name=\'{}\')'.format(
            self.__class__.__name__,
            self.id,
            self.name
        )

    def send_test_notification(self) -> str:
        """Sends a test notification to this webhook and returns send report"""
        hook = dhooks_lite.Webhook(self.url)
        response = hook.execute(
            _(
                'This is a test notification from %s.\n'
                'The webhook appears to be correctly configured.'
            ) % __title__,
            wait_for_response=True
        )
        if response.status_ok:
            send_report_json = json.dumps(
                response.content, indent=4, sort_keys=True
            )
        else:
            send_report_json = 'HTTP status code {}'.format(
                response.status_code
            )
        return send_report_json


class EveEntity(models.Model):
    """An EVE entity like a character or an alliance"""

    CATEGORY_CHARACTER = 1
    CATEGORY_CORPORATION = 2
    CATEGORY_ALLIANCE = 3
    CATEGORY_FACTION = 4
    CATEGORY_OTHER = 5

    CATEGORY_CHOICES = [
        (CATEGORY_CHARACTER, 'character'),
        (CATEGORY_CORPORATION, 'corporation'),
        (CATEGORY_ALLIANCE, 'alliance'),
        (CATEGORY_FACTION, 'faction'),
        (CATEGORY_OTHER, 'other'),
    ]

    id = models.PositiveIntegerField(
        primary_key=True, help_text='Eve Online ID'
    )
    category = models.IntegerField(choices=CATEGORY_CHOICES)
    name = models.CharField(
        max_length=255,
        null=True,
        default=None,
        blank=True
    )

    objects = EveEntityManager()

    def __str__(self):
        return str(self.name)

    def __repr__(self):
        return '{}(id={}, category=\'{}\', name=\'{}\')'.format(
            self.__class__.__name__,
            self.id,
            self.get_category_display(),
            self.name
        )

    def profile_url(self) -> str:
        """returns link to website with profile info about this entity"""        
        if self.category == self.CATEGORY_CORPORATION:
            url = dotlan.corporation_url(self.name)
        elif self.category == self.CATEGORY_ALLIANCE:
            url = dotlan.alliance_url(self.name)        
        else:
            url = ''
        return url

    @classmethod
    def get_matching_entity_category(cls, type_name) -> int:
        """returns category for given ESI name"""
        match = None
        for x in cls.CATEGORY_CHOICES:
            if type_name == x[1]:
                match = x
                break
        return match[0] if match else cls.CATEGORY_OTHER


class Notification(models.Model):
    """An EVE Online notification about structures"""

    # embed colors
    EMBED_COLOR_INFO = 0x5bc0de
    EMBED_COLOR_SUCCESS = 0x5cb85c
    EMBED_COLOR_WARNING = 0xf0ad4e
    EMBED_COLOR_DANGER = 0xd9534f

    HTTP_CODE_TOO_MANY_REQUESTS = 429

    # event type structure map
    MAP_CAMPAIGN_EVENT_2_TYPE_ID = {
        1: EveType.EVE_TYPE_ID_TCU,
        2: EveType.EVE_TYPE_ID_IHUB
    }

    MAP_TYPE_ID_2_TIMER_STRUCTURE_NAME = {
        2233: 'POCO',
        32226: 'TCU',
        32458: 'I-HUB'
    }

    notification_id = models.BigIntegerField( 
        validators=[MinValueValidator(0)]
    )
    owner = models.ForeignKey(
        'Owner',
        on_delete=models.CASCADE,
        help_text='Corporation that received this notification'
    )
    sender = models.ForeignKey(EveEntity, on_delete=models.CASCADE)
    timestamp = models.DateTimeField()
    notification_type = models.IntegerField(choices=NTYPE_CHOICES)
    text = models.TextField(
        null=True,
        default=None,
        blank=True,
        help_text='Notification details in YAML'
    )
    is_read = models.BooleanField(
        null=True,
        default=None,
        blank=True,
        help_text='True when this notification has read in the eve client'
    )
    is_sent = models.BooleanField(
        default=False,
        blank=True,
        help_text='True when this notification has been forwarded to Discord'
    )
    is_timer_added = models.BooleanField(
        null=True,
        default=None,
        blank=True,
        help_text='True when a timer has been added for this notification'
    )
    last_updated = models.DateTimeField(
        help_text='Date when this notification has last been updated from ESI'
    )
    created = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        help_text='Date when this notification was first received from ESI'
    )

    class Meta:
        unique_together = (('notification_id', 'owner'),)

    def __str__(self):
        return str(self.notification_id)

    def __repr__(self):
        return '%s(notification_id=%d, owner=\'%s\', notification_type=\'%s\')' % (
            self.__class__.__name__,            
            self.notification_id,
            self.owner,
            self.get_notification_type_display(),
        )

    @property
    def is_alliance_level(self):
        """whether this is an alliance level notification"""
        return self.notification_type in NTYPE_FOR_ALLIANCE_LEVEL
    
    @classmethod
    def get_all_types(cls):
        """returns a set with all supported notification types"""
        return {x[0] for x in NTYPE_CHOICES}

    @classmethod
    def get_all_type_names(cls):
        """returns a set with names of all supported notification types"""
        return {x[1] for x in NTYPE_CHOICES}

    @classmethod
    def get_types_for_timerboard(cls):
        """returns set of types relevant for the timerboard"""
        return _NTYPE_RELEVANT_FOR_TIMERBOARD

    @classmethod
    def get_matching_notification_type(cls, type_name) -> int:
        """returns matching notification type for given name or None"""
        match = None
        for x in NTYPE_CHOICES:
            if type_name == x[1]:
                match = x
                break
        
        return match[0] if match else None

    def get_parsed_text(self) -> dict:
        """returns the notifications's text as dict"""
        return yaml.safe_load(self.text)

    def is_npc_attacking(self):
        """ whether this notification is about a NPC attacking"""
        result = False
        if self.notification_type in [
            NTYPE_ORBITAL_ATTACKED,
            NTYPE_STRUCTURE_UNDER_ATTACK
        ]:
            parsed_text = self.get_parsed_text()
            corporation_id = None
            if self.notification_type == NTYPE_STRUCTURE_UNDER_ATTACK:
                if ('corpLinkData' in parsed_text
                    and len(parsed_text['corpLinkData']) >= 3
                ):
                    corporation_id = int(parsed_text['corpLinkData'][2])

            if self.notification_type == NTYPE_ORBITAL_ATTACKED:
                if 'aggressorCorpID' in parsed_text:
                    corporation_id = int(parsed_text['aggressorCorpID'])

            if 1000000 <= corporation_id <= 2000000:
                result = True

        return result

    def filter_for_npc_attacks(self):
        """true when notification to be filtered out due to npc attacks"""
        return not STRUCTURES_REPORT_NPC_ATTACKS and self.is_npc_attacking()

    def filter_for_alliance_level(self):
        """true when notification to be filtered out due to alliance level"""
        return self.is_alliance_level and not self.owner.is_alliance_main

    def send_to_webhook(self, webhook: Webhook) -> bool:
        """sends this notification to the configured webhook
        returns True if successful, else False
        """        
        add_prefix = make_logger_prefix(
            'notification:{}'.format(self.notification_id)
        )
        logger.info(add_prefix('Trying to sent to webhook: %s' % webhook))
        if self.is_alliance_level:
            avatar_url = self.owner.corporation.alliance.logo_url()
            ticker = self.owner.corporation.alliance.alliance_ticker
        else:
            avatar_url = self.owner.corporation.logo_url()
            ticker = self.owner.corporation.corporation_ticker

        username = gettext('%(ticker)s Notification') % {'ticker': ticker}
        hook = dhooks_lite.Webhook(
            webhook.url, username=username, avatar_url=avatar_url
        )        
        success = False
        try:
            embed = self._generate_embed(webhook.language_code)
        except Exception as ex:
            logger.warning(add_prefix('Failed to generate embed: %s' % ex))
            raise ex
        else:
            if embed.color == self.EMBED_COLOR_DANGER:
                content = '@everyone'
            elif embed.color == self.EMBED_COLOR_WARNING:
                content = '@here'
            else:
                content = None

            success = self._execute_webhook(hook, content, embed, add_prefix)
        return success

    @classmethod
    def _ldap_datetime_2_dt(cls, ldap_dt: int) -> datetime:
        """converts ldap time to datatime"""
        return pytz.utc.localize(datetime.utcfromtimestamp(
            (ldap_dt / 10000000) - 11644473600
        ))

    @classmethod
    def _ldap_timedelta_2_timedelta(cls, ldap_td: int) -> timedelta:
        """converts a ldap timedelta into a dt timedelta"""
        return timedelta(microseconds=ldap_td / 10)

    def _generate_embed(self, language_code: str) -> dhooks_lite.Embed:
        """generates a Discord embed for this notification"""

        logger.info('Creating embed with language = %s' % language_code)
        parsed_text = self.get_parsed_text()
        
        with translation.override(language_code):
            if self.notification_type in [
                NTYPE_STRUCTURE_FUEL_ALERT,
                NTYPE_STRUCTURE_SERVICES_OFFLINE,
                NTYPE_STRUCTURE_WENT_LOW_POWER,
                NTYPE_STRUCTURE_WENT_HIGH_POWER,
                NTYPE_STRUCTURE_UNANCHORING,
                NTYPE_STRUCTURE_UNDER_ATTACK,
                NTYPE_STRUCTURE_LOST_SHIELD,
                NTYPE_STRUCTURE_LOST_ARMOR,
                NTYPE_STRUCTURE_DESTROYED,
                NTYPE_STRUCTURE_ONLINE
            ]:
                title, description, color, thumbnail = \
                    self._gen_embed_structures_1(parsed_text)

            elif self.notification_type in [
                NTYPE_OWNERSHIP_TRANSFERRED,
                NTYPE_STRUCTURE_ANCHORING
            ]:
                title, description, color, thumbnail = \
                    self._gen_embed_structures_2(parsed_text)

            elif self.notification_type in [
                NTYPE_MOONS_AUTOMATIC_FRACTURE,
                NTYPE_MOONS_EXTRACTION_CANCELED,
                NTYPE_MOONS_EXTRACTION_FINISHED,
                NTYPE_MOONS_EXTRACTION_STARTED,
                NTYPE_MOONS_LASER_FIRED
            ]:
                title, description, color, thumbnail = \
                    self._gen_embed_moons(parsed_text)

            elif self.notification_type in [
                NTYPE_ORBITAL_ATTACKED,
                NTYPE_ORBITAL_REINFORCED,
            ]:
                title, description, color, thumbnail = \
                    self._gen_embed_pocos(parsed_text)

            elif self.notification_type in [
                NTYPE_TOWER_ALERT_MSG,
                NTYPE_TOWER_RESOURCE_ALERT_MSG,
            ]:
                title, description, color, thumbnail = \
                    self._gen_embed_poses(parsed_text)

            elif self.notification_type in [
                NTYPE_SOV_ENTOSIS_CAPTURE_STARTED,
                NTYPE_SOV_COMMAND_NODE_EVENT_STARTED,
                NTYPE_SOV_ALL_CLAIM_ACQUIRED_MSG,
                NTYPE_SOV_STRUCTURE_REINFORCED,
                NTYPE_SOV_STRUCTURE_DESTROYED
            ]:
                title, description, color, thumbnail = \
                    self._gen_embed_sov(parsed_text)

            else:
                raise NotImplementedError('type: {}'.format(
                    self.notification_type
                ))

        if STRUCTURES_DEVELOPER_MODE:
            footer = dhooks_lite.Footer(self.notification_id)
        else:
            footer = None

        return dhooks_lite.Embed(
            title=title,
            description=description,
            color=color,
            thumbnail=thumbnail,
            timestamp=self.timestamp,
            footer=footer
        )
    
    def _gen_embed_structures_1(self, parsed_text: dict) -> tuple:
        
        try:
            my_structure = Structure.objects.get(id=parsed_text['structureID'])
            structure_name = my_structure.name
            structure_type = my_structure.eve_type
            structure_solar_system = my_structure.eve_solar_system
        except Structure.DoesNotExist:
            my_structure = None
            structure_name = gettext('(unknown)')
            structure_type, _ = EveType.objects.get_or_create_esi(
                parsed_text['structureTypeID']
            )
            structure_solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
                parsed_text['solarsystemID']
            )

        description = gettext(
            'The %(structure_type)s %(structure_name)s in %(solar_system)s '
        ) % {
            'structure_type': structure_type.name_localized,
            'structure_name': '**%s**' % structure_name,
            'solar_system': self._gen_solar_system_text(structure_solar_system)
        }
        if self.notification_type == NTYPE_STRUCTURE_ONLINE:
            title = gettext('Structure online')
            description += gettext('is now online.')
            color = self.EMBED_COLOR_SUCCESS

        if self.notification_type == NTYPE_STRUCTURE_FUEL_ALERT:
            title = gettext('Structure fuel alert')
            description += gettext('has less then 24hrs fuel left.')
            color = self.EMBED_COLOR_WARNING

        elif self.notification_type == NTYPE_STRUCTURE_SERVICES_OFFLINE:
            title = gettext('Structure services off-line')
            description += gettext('has all services off-lined.')
            if my_structure and my_structure.structureservice_set.count() > 0:
                qs = my_structure.structureservice_set.all().order_by('name')
                services_list = '\n'.join([x.name for x in qs])
                description += '\n*{}*'.format(
                    services_list
                )

            color = self.EMBED_COLOR_DANGER

        elif self.notification_type == NTYPE_STRUCTURE_WENT_LOW_POWER:
            title = gettext('Structure low power')
            description += gettext('went to low power mode.')
            color = self.EMBED_COLOR_WARNING

        elif self.notification_type == NTYPE_STRUCTURE_WENT_HIGH_POWER:
            title = gettext('Structure full power')
            description += gettext('went to full power mode.')
            color = self.EMBED_COLOR_SUCCESS

        elif self.notification_type == NTYPE_STRUCTURE_UNANCHORING:
            title = gettext('Structure un-anchoring')
            unanchored_at = self.timestamp \
                + self._ldap_timedelta_2_timedelta(parsed_text['timeLeft'])
            description += gettext(
                'has started un-anchoring. '
                'It will be fully un-anchored at: %s'
            ) % unanchored_at.strftime(DATETIME_FORMAT)
            color = self.EMBED_COLOR_INFO

        elif self.notification_type == NTYPE_STRUCTURE_UNDER_ATTACK:
            title = gettext('Structure under attack')
            description += gettext('is under attack by %s') \
                % self._get_attacker_link(parsed_text)
            color = self.EMBED_COLOR_DANGER

        elif self.notification_type == NTYPE_STRUCTURE_LOST_SHIELD:
            title = gettext('Structure lost shield')
            timer_ends_at = self.timestamp \
                + self._ldap_timedelta_2_timedelta(parsed_text['timeLeft'])
            description += gettext(
                'has lost its shields. Armor timer end at: %s'
            ) % timer_ends_at.strftime(DATETIME_FORMAT)
            color = self.EMBED_COLOR_DANGER

        elif self.notification_type == NTYPE_STRUCTURE_LOST_ARMOR:
            title = gettext('Structure lost armor')
            timer_ends_at = self.timestamp \
                + self._ldap_timedelta_2_timedelta(parsed_text['timeLeft'])
            description += gettext(
                'has lost its armor. Hull timer end at: %s'
            ) % timer_ends_at.strftime(DATETIME_FORMAT)
            color = self.EMBED_COLOR_DANGER

        elif self.notification_type == NTYPE_STRUCTURE_DESTROYED:
            title = gettext('Structure destroyed')
            description += gettext('has been destroyed.')
            color = self.EMBED_COLOR_DANGER
        
        thumbnail = dhooks_lite.Thumbnail(structure_type.icon_url())
        return title, description, color, thumbnail
    
    def _gen_embed_structures_2(self, parsed_text: dict) -> tuple:
        structure_type, _ = EveType.objects.get_or_create_esi(
            parsed_text['structureTypeID']
        )        
        if self.notification_type == NTYPE_OWNERSHIP_TRANSFERRED:            
            solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
                parsed_text['solarSystemID']
            )
            description = gettext(
                'The %(structure_type)s %(structure_name)s '
                'in %(solar_system)s '
            ) % {
                'structure_type': structure_type.name,
                'structure_name': '**%s**' % parsed_text['structureName'],
                'solar_system': self._gen_solar_system_text(solar_system)
            }
            from_corporation, _ = \
                EveEntity.objects.get_or_create_esi(
                    parsed_text['oldOwnerCorpID']
                )
            to_corporation, _ = \
                EveEntity.objects.get_or_create_esi(
                    parsed_text['newOwnerCorpID']
                )
            character, _ = \
                EveEntity.objects.get_or_create_esi(
                    parsed_text['charID']
                )
            description += gettext(
                'has been transferred from %(from_corporation)s '
                'to %(to_corporation)s by %(character)s.'
            ) % {
                'from_corporation': self._gen_corporation_link(
                    from_corporation.name
                ),
                'to_corporation': self._gen_corporation_link(
                    to_corporation.name
                ),
                'character': character.name
            }
            title = gettext('Ownership transferred')
            color = self.EMBED_COLOR_INFO

        elif self.notification_type == NTYPE_STRUCTURE_ANCHORING:            
            solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
                parsed_text['solarsystemID']
            )
            description = gettext(
                '%(structure_type)s has started anchoring '
                'in %(solar_system)s. '
            ) % {
                'structure_type': structure_type.name_localized,
                'solar_system': self._gen_solar_system_text(solar_system)
            }
            if not solar_system.is_null_sec:
                unanchored_at = self.timestamp + timedelta(hours=24)
                description += 'The anchoring timer ends at: {}'.format(
                    unanchored_at.strftime(DATETIME_FORMAT)
                )
            title = gettext('Structure anchoring')
            color = self.EMBED_COLOR_INFO

        thumbnail = dhooks_lite.Thumbnail(structure_type.icon_url())
        return title, description, color, thumbnail
    
    def _gen_embed_moons(self, parsed_text: dict) -> tuple:
        moon, _ = EveMoon.objects.get_or_create_esi(parsed_text['moonID'])        
        solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            parsed_text['solarSystemID']
        )
        solar_system_link = self._gen_solar_system_text(solar_system)
        structure_name = parsed_text['structureName']
        if self.notification_type == NTYPE_MOONS_EXTRACTION_STARTED:
            started_by, _ = EveEntity.objects.get_or_create_esi(
                parsed_text['startedBy']
            )
            ready_time = self._ldap_datetime_2_dt(parsed_text['readyTime'])
            auto_time = self._ldap_datetime_2_dt(parsed_text['autoTime'])
            title = gettext('Moon mining extraction started')
            description = gettext(
                'A moon mining extraction has been started '
                'for %(structure_name)s at %(moon)s in %(solar_system)s. '
                'Extraction was started by %(character)s.\n'
                'The chunk will be ready on location at %(ready_time)s, '
                'and will autofracture on %(auto_time)s.\n'
            ) % {
                'structure_name': '**%s**' % structure_name,
                'moon': moon.name_localized,
                'solar_system': solar_system_link,
                'character': started_by,
                'ready_time': ready_time.strftime(DATETIME_FORMAT),
                'auto_time': auto_time.strftime(DATETIME_FORMAT)
            }                
            color = self.EMBED_COLOR_INFO

        elif (self.notification_type == NTYPE_MOONS_EXTRACTION_FINISHED):
            auto_time = self._ldap_datetime_2_dt(parsed_text['autoTime'])
            title = gettext('Extraction finished')
            description = gettext(
                'The extraction for %(structure_name)s at %(moon)s '
                'in %(solar_system)s is finished and the chunk is ready '
                'to be shot at.\n'
                'The chunk will automatically fracture on %(auto_time)s.'
            ) % {
                'structure_name': '**%s**' % structure_name,
                'moon': moon.name_localized,
                'solar_system': solar_system_link,
                'auto_time': auto_time.strftime(DATETIME_FORMAT)
            }                
            color = self.EMBED_COLOR_INFO

        elif self.notification_type == NTYPE_MOONS_AUTOMATIC_FRACTURE:
            title = gettext('Automatic Fracture')
            description = gettext(
                'The moondrill fitted to %(structure_name)s at %(moon)s'
                ' in %(solar_system)s has automatically been fired '
                'and the moon products are ready to be harvested.\n'
            ) % {
                'structure_name': '**%s**' % structure_name,
                'moon': moon.name_localized,
                'solar_system': solar_system_link
            }                
            color = self.EMBED_COLOR_SUCCESS

        elif (self.notification_type == NTYPE_MOONS_EXTRACTION_CANCELED):
            if parsed_text['cancelledBy']:
                cancelled_by, _ = EveEntity.objects.get_or_create_esi(
                    parsed_text['cancelledBy']
                )
            else:
                cancelled_by = gettext('(unknown)')
            title = gettext('Extraction cancelled')
            description = gettext(
                'An ongoing extraction for %(structure_name)s at %(moon)s '
                'in %(solar_system)s has been cancelled by %(character)s.'
            ) % {
                'structure_name': '**%s**' % structure_name,
                'moon': moon.name_localized,
                'solar_system': solar_system_link,
                'character': cancelled_by
            }
        
            color = self.EMBED_COLOR_WARNING

        elif self.notification_type == NTYPE_MOONS_LASER_FIRED:
            fired_by, _ = EveEntity.objects.get_or_create_esi(
                parsed_text['firedBy']
            )
            title = gettext('Moondrill fired')
            description = gettext(
                'The moondrill fitted to %(structure_name)s at %(moon)s '
                'in %(solar_system)s has been fired by %(character)s '
                'and the moon products are ready to be harvested.'
            ) % {
                'structure_name': '**%s**' % structure_name,
                'moon': moon.name_localized,
                'solar_system': solar_system_link,
                'character': fired_by
            }
            color = self.EMBED_COLOR_SUCCESS

        structure_type, _ = EveType.objects.get_or_create_esi(
            parsed_text['structureTypeID']
        )
        thumbnail = dhooks_lite.Thumbnail(structure_type.icon_url())
        return title, description, color, thumbnail

    def _gen_embed_pocos(self, parsed_text: dict) -> tuple:        
        planet, _ = EvePlanet.objects.get_or_create_esi(
            parsed_text['planetID']
        )
        structure_type, _ = EveType.objects.get_or_create_esi(
            EveType.EVE_TYPE_ID_POCO
        )        
        solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            parsed_text['solarSystemID']
        )
        solar_system_link = self._gen_solar_system_text(solar_system)
        aggressor_link = self._get_aggressor_link(parsed_text)

        if self.notification_type == NTYPE_ORBITAL_ATTACKED:
            title = gettext('Orbital under attack')
            description = gettext(
                'The %(structure_type)s at %(planet)s in %(solar_system)s '
                'is under attack by %(aggressor)s.'
            ) % {
                'structure_type': structure_type.name_localized,
                'planet': planet.name_localized,
                'solar_system': solar_system_link,
                'aggressor': aggressor_link
            }
            color = self.EMBED_COLOR_WARNING

        elif self.notification_type == NTYPE_ORBITAL_REINFORCED:
            reinforce_exit_time = \
                self._ldap_datetime_2_dt(parsed_text['reinforceExitTime'])
            title = gettext('Orbital reinforced')
            description = gettext(
                'The %(structure_type)s at %(planet)s in %(solar_system)s '
                'has been reinforced by %(aggressor)s '
                'and will come out at: %(date)s.'
            ) % {
                'structure_type': structure_type.name_localized,
                'planet': planet.name_localized,
                'solar_system': solar_system_link,
                'aggressor': aggressor_link,
                'date': reinforce_exit_time.strftime(DATETIME_FORMAT)
            }
            color = self.EMBED_COLOR_DANGER
        
        thumbnail = dhooks_lite.Thumbnail(structure_type.icon_url())
        return title, description, color, thumbnail
    
    def _gen_embed_poses(self, parsed_text: dict) -> tuple:        
        eve_moon, _ = EveMoon.objects.get_or_create_esi(
            parsed_text['moonID']
        )
        structure_type, _ = EveType.objects.get_or_create_esi(
            parsed_text['typeID']
        )        
        solar_system_link = self._gen_solar_system_text(
            eve_moon.eve_solar_system
        )
        qs_structures = Structure.objects.filter(eve_moon=eve_moon)
        if qs_structures.exists():
            structure_name = qs_structures.first().name
        else:
            structure_name = structure_type.name_localized

        if self.notification_type == NTYPE_TOWER_ALERT_MSG:
            aggressor_link = self._get_aggressor_link(parsed_text)
            damage_labels = [
                ('shield', gettext('shield')), 
                ('armor', gettext('armor')), 
                ('hull', gettext('hull')),                    
            ]
            damage_parts = list()
            for prop in damage_labels:
                prop_yaml = prop[0] + 'Value'
                if prop_yaml in parsed_text:
                    damage_parts.append(
                        '{}: {:.0f}%'.format(
                            prop[1],
                            parsed_text[prop_yaml] * 100
                        )
                    )
            damage_text = ' | '.join(damage_parts)
            title = gettext('Starbase under attack')
            description = gettext(
                'The starbase %(structure_name)s at %(moon)s '
                'in %(solar_system)s is under attack by %(aggressor)s.\n'
                '%(damage_text)s'
            ) % {
                'structure_name': '**%s**' % structure_name,
                'moon': eve_moon.name_localized,
                'solar_system': solar_system_link,
                'aggressor': aggressor_link,
                'damage_text': damage_text,
            }
            color = self.EMBED_COLOR_WARNING

        elif self.notification_type == NTYPE_TOWER_RESOURCE_ALERT_MSG:
            quantity = parsed_text['wants'][0]['quantity']
            title = gettext('Starbase low on fuel')
            description = gettext(
                'The starbase %(structure_name)s at %(moon)s '
                'in %(solar_system)s is low on fuel. '
                'It has %(quantity)d fuel blocks left.'
            ) % {
                'structure_name': '**%s**' % structure_name,
                'moon': eve_moon.name_localized,
                'solar_system': solar_system_link,
                'quantity': quantity
            }
            color = self.EMBED_COLOR_WARNING
        
        thumbnail = dhooks_lite.Thumbnail(structure_type.icon_url())
        return title, description, color, thumbnail
    
    def _gen_embed_sov(self, parsed_text: dict) -> tuple:        
        solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            parsed_text['solarSystemID']
        )
        solar_system_link = self._gen_solar_system_text(solar_system)

        if 'structureTypeID' in parsed_text:
            structure_type_id = parsed_text['structureTypeID']
        elif 'campaignEventType' in parsed_text:
            structure_type_id = self._get_type_id_from_event_type(
                parsed_text['campaignEventType']
            )
        else:
            structure_type_id = EveType.EVE_TYPE_ID_TCU

        structure_type, _ = EveType.objects.get_or_create_esi(
            structure_type_id
        )
        structure_type_name = structure_type.name_localized        
        sov_owner_link = self._gen_alliance_link(
            self.owner.corporation.alliance.alliance_name
        )
        if self.notification_type == NTYPE_SOV_ENTOSIS_CAPTURE_STARTED:
            title = gettext(
                '%(structure_type)s in %(solar_system)s is being captured'
            ) % {
                'structure_type': '**%s**' % structure_type_name,
                'solar_system': solar_system.name_localized
            }
            description = gettext(
                'A capsuleer has started to influence the %(type)s '
                'in %(solar_system)s belonging to %(owner)s '
                'with an Entosis Link.'
            ) % {
                'type': structure_type_name,
                'solar_system': solar_system_link,
                'owner': sov_owner_link
            }
            color = self.EMBED_COLOR_WARNING

        elif (self.notification_type == NTYPE_SOV_COMMAND_NODE_EVENT_STARTED):
            title = gettext(
                'Command nodes for %(structure_type)s in %(solar_system)s '
                'have begun to decloak'
            ) % {
                'structure_type': '**%s**' % structure_type_name,
                'solar_system': solar_system.name_localized
            }
            description = gettext(
                'Command nodes for %(structure_type)s in %(solar_system)s '
                'can now be found throughout '
                'the %(constellation)s constellation'
            ) % {
                'structure_type': '**%s**' % structure_type_name,
                'solar_system': solar_system_link,
                'constellation': solar_system.eve_constellation.name_localized
            }
            color = self.EMBED_COLOR_WARNING

        elif self.notification_type == NTYPE_SOV_ALL_CLAIM_ACQUIRED_MSG:
            alliance, _ = EveEntity.objects.get_or_create_esi(
                parsed_text['allianceID']
            )
            corporation, _ = EveEntity.objects.get_or_create_esi(
                parsed_text['corpID']
            )
            title = gettext(
                'DED Sovereignty claim acknowledgment: %s'
            ) % solar_system.name_localized
            
            description = gettext(
                'DED now officially acknowledges that your '
                'member corporation %(corporation)s has claimed '
                'sovereignty on behalf of %(alliance)s in %(solar_system)s.'
            ) % {
                'corporation': self._gen_corporation_link(corporation.name),
                'alliance': self._gen_alliance_link(alliance.name),
                'solar_system': solar_system_link
            }
            color = self.EMBED_COLOR_SUCCESS

        elif self.notification_type == NTYPE_SOV_STRUCTURE_REINFORCED:
            timer_starts = self._ldap_datetime_2_dt(
                parsed_text['decloakTime']
            )
            title = gettext(
                '%(structure_type)s in %(solar_system)s '
                'has entered reinforced mode'
            ) % {
                'structure_type': '**%s**' % structure_type_name,
                'solar_system': solar_system.name_localized
            }
            description = gettext(
                'The %(structure_type)s in %(solar_system)s belonging '
                'to %(owner)s has been reinforced by '
                'hostile forces and command nodes '
                'will begin decloaking at %(date)s'
            ) % {
                'structure_type': '**%s**' % structure_type_name,
                'solar_system': solar_system_link,
                'owner': sov_owner_link,
                'date': timer_starts.strftime(DATETIME_FORMAT)
            }
            color = self.EMBED_COLOR_DANGER

        elif self.notification_type == NTYPE_SOV_STRUCTURE_DESTROYED:
            title = gettext(
                '%(structure_type)s in %(solar_system)s has been destroyed'
            ) % {
                'structure_type': '**%s**' % structure_type_name,
                'solar_system': solar_system.name_localized
            }
            description = gettext(
                'The command nodes for %(structure_type)s '
                'in %(solar_system)s belonging to %(owner)s have been '
                'destroyed by hostile forces.'
            ) % {
                'structure_type': '**%s**' % structure_type_name,
                'solar_system': solar_system_link,
                'owner': sov_owner_link
            }
            color = self.EMBED_COLOR_DANGER
        
        thumbnail = dhooks_lite.Thumbnail(structure_type.icon_url())
        return title, description, color, thumbnail

    @classmethod
    def _gen_solar_system_text(cls, solar_system: EveSolarSystem) -> str:
        text = '[{}]({}) ({})'.format(
            solar_system.name_localized,
            dotlan.solar_system_url(solar_system.name),
            solar_system.eve_constellation.eve_region.name_localized
        )
        return text

    @classmethod
    def _gen_alliance_link(cls, alliance_name):
        return '[{}]({})'.format(
            alliance_name, dotlan.alliance_url(alliance_name)
        )

    @classmethod
    def _gen_corporation_link(cls, corporation_name):
        return '[{}]({})'.format(
            corporation_name, dotlan.corporation_url(corporation_name)
        )

    @classmethod
    def _get_attacker_link(cls, parsed_text):
        """returns the attacker link from a parsed_text
        For Upwell structures only
        """
        if "allianceName" in parsed_text:
            name = cls._gen_alliance_link(parsed_text['allianceName'])
        elif "corpName" in parsed_text:
            name = cls._gen_corporation_link(parsed_text['corpName'])
        else:
            name = "(unknown)"

        return name

    @classmethod
    def _get_aggressor_link(cls, parsed_text: dict) -> str:
        """returns the aggressor link from a parsed_text
        for POS and POCOs only
        """
        if 'aggressorAllianceID' in parsed_text:
            key = 'aggressorAllianceID'
        elif 'aggressorCorpID' in parsed_text:
            key = 'aggressorCorpID'
        elif 'aggressorID' in parsed_text:
            key = 'aggressorID'
        else:
            return '(Unknown aggressor)'

        entity, _ = EveEntity.objects.get_or_create_esi(parsed_text[key])
        return '[{}]({})'.format(entity.name, entity.profile_url())

    @classmethod
    def _get_type_id_from_event_type(cls, event_type: int) -> int:
        if event_type in cls.MAP_CAMPAIGN_EVENT_2_TYPE_ID:
            return cls.MAP_CAMPAIGN_EVENT_2_TYPE_ID[event_type]
        else:
            return None
    
    def _execute_webhook(
        self, hook, content, embed, add_prefix
    ) -> bool:
        """executes webhook for sending the message, will retry on errors
        
        Sets this notification as "sent" if successful
        
        returns True/False on success
        """
        success = False
        max_retries = STRUCTURES_NOTIFICATION_MAX_RETRIES
        for retry_count in range(max_retries + 1):
            if retry_count > 0:
                logger.warn(add_prefix('Retry {} / {}'.format(
                    retry_count, max_retries
                )))
            try:
                res = hook.execute(
                    content=content, embeds=[embed], wait_for_response=True
                )
                if res.status_ok:
                    self.is_sent = True
                    self.save()
                    success = True
                    break

                elif res.status_code == self.HTTP_CODE_TOO_MANY_REQUESTS:
                    if 'retry_after' in res.content:
                        retry_after = res.content['retry_after'] / 1000
                    else:
                        retry_after = STRUCTURES_NOTIFICATION_WAIT_SEC
                    logger.warn(add_prefix(
                        'rate limited - retry after %d secs' % retry_after
                    ))
                    sleep(retry_after)

                else:
                    logger.warn(add_prefix(
                        'HTTP error {} while trying '
                        'to send notifications'.format(res.status_code) 
                    ))
                    if retry_count < max_retries + 1:
                        sleep(STRUCTURES_NOTIFICATION_WAIT_SEC)

            except Exception as ex:
                logger.warn(add_prefix(
                    'Unexpected issue when trying to send message: %s' % ex
                ))
                if settings.DEBUG:
                    raise ex
                else:
                    break

        return success

    def process_for_timerboard(self, token: Token = None) -> bool:
        """add/removes a timer related to this notification for some types
        returns True when a timer was processed, else False
        """
        success = False
        if self.notification_type in _NTYPE_RELEVANT_FOR_TIMERBOARD:
            parsed_text = self.get_parsed_text()
            try:
                with translation.override(STRUCTURES_DEFAULT_LANGUAGE):
                    if self.notification_type in [
                        NTYPE_STRUCTURE_LOST_ARMOR,
                        NTYPE_STRUCTURE_LOST_SHIELD,
                    ]:
                        timer = self._gen_timer_structure_reinforcement(
                            parsed_text, token
                        )
                    elif self.notification_type == NTYPE_STRUCTURE_ANCHORING:
                        timer = self._gen_timer_structure_anchoring(parsed_text)
                    elif self.notification_type == NTYPE_SOV_STRUCTURE_REINFORCED:
                        timer = self._gen_timer_sov_reinforcements(parsed_text)
                    elif self.notification_type == NTYPE_ORBITAL_REINFORCED:
                        timer = self._gen_timer_orbital_reinforcements(parsed_text)
                    elif self.notification_type in [
                        NTYPE_MOONS_EXTRACTION_STARTED,
                        NTYPE_MOONS_EXTRACTION_CANCELED
                    ]:
                        if not STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED:
                            timer = None
                        else:
                            timer = self._gen_timer_moon_extraction(parsed_text)
                    else:
                        raise NotImplementedError()

                if timer:
                    timer.save()
                    logger.info(
                        '{}: added timer related notification'.format(
                            self.notification_id
                        )
                    )
                    self.is_timer_added = True
                    self.save()
                    success = True

            except Exception as ex:
                logger.exception(
                    '{}: Failed to add timer from notification: {}'.format(
                        self.notification_id,
                        ex
                    )
                )
                if settings.DEBUG:
                    raise ex

        return success
    
    def _gen_timer_structure_reinforcement(self, parsed_text, token):
        """generate timer for structure reinforcements"""
        structure_obj, _ = Structure.objects.get_or_create_esi(
            parsed_text['structureID'], token
        )
        eve_time = self.timestamp \
            + self._ldap_timedelta_2_timedelta(
                parsed_text['timeLeft']
            )
        if self.notification_type == NTYPE_STRUCTURE_LOST_SHIELD:
            details = gettext('Armor timer')
        elif self.notification_type == NTYPE_STRUCTURE_LOST_ARMOR:
            details = gettext('Final timer')

        return Timer(
            details=details,
            system=structure_obj.eve_solar_system.name,
            planet_moon='',
            structure=structure_obj.eve_type.name,
            objective='Friendly',
            eve_time=eve_time,
            eve_corp=self.owner.corporation,
            corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
        )
    
    def _gen_timer_structure_anchoring(self, parsed_text):
        """generate timer for structure anchoring"""
        structure_type, _ = EveType.objects.get_or_create_esi(
            parsed_text['structureTypeID']
        )
        solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            parsed_text['solarsystemID']
        )
        if not solar_system.is_null_sec:
            eve_time = self.timestamp + timedelta(hours=24)
            timer = Timer(
                details=gettext('Anchor timer'),
                system=solar_system.name,
                planet_moon='',
                structure=structure_type.name,
                objective='Friendly',
                eve_time=eve_time,
                eve_corp=self.owner.corporation,
                corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
            )
        else:
            timer = None
        
        return timer
    
    def _gen_timer_sov_reinforcements(self, parsed_text):
        """generate timer for sov reinforcements"""
        solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            parsed_text['solarSystemID']
        )
        event_type = parsed_text['campaignEventType']
        if event_type in self.MAP_CAMPAIGN_EVENT_2_TYPE_ID:
            structure_type_name = \
                self.MAP_TYPE_ID_2_TIMER_STRUCTURE_NAME[
                    self.MAP_CAMPAIGN_EVENT_2_TYPE_ID[event_type]
                ]
        else:
            structure_type_name = 'Other'

        eve_time = self._ldap_datetime_2_dt(parsed_text['decloakTime'])
        return Timer(
            details=gettext('Sov timer'),
            system=solar_system.name,
            planet_moon='',
            structure=structure_type_name,
            objective='Friendly',
            eve_time=eve_time,
            eve_corp=self.owner.corporation,
            corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
        )
    
    def _gen_timer_orbital_reinforcements(self, parsed_text):
        """generate timer for orbital reinforcements"""
        solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            parsed_text['solarSystemID']
        )
        planet, _ = EvePlanet.objects.get_or_create_esi(
            parsed_text['planetID']
        )
        eve_time = self._ldap_datetime_2_dt(
            parsed_text['reinforceExitTime']
        )
        return Timer(
            details=gettext('Final timer'),
            system=solar_system.name,
            planet_moon=planet.name,
            structure='POCO',
            objective='Friendly',
            eve_time=eve_time,
            eve_corp=self.owner.corporation,
            corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
        )
    
    def _gen_timer_moon_extraction(self, parsed_text):
        """generate timer for moon mining extractions"""
        solar_system, _ = \
            EveSolarSystem.objects.get_or_create_esi(
                parsed_text['solarSystemID']
            )
        moon, _ = EveMoon.objects.get_or_create_esi(
            parsed_text['moonID']
        )
        if 'readyTime' in parsed_text:
            eve_time = self._ldap_datetime_2_dt(
                parsed_text['readyTime']
            )
        else:
            eve_time = None
        details = gettext('Extraction ready')
        system = solar_system.name
        planet_moon = moon.name
        structure_type_name = 'Moon Mining Cycle'
        objective = 'Friendly'

        if self.notification_type == NTYPE_MOONS_EXTRACTION_STARTED:
            timer = Timer(
                details=details,
                system=system,
                planet_moon=planet_moon,
                structure=structure_type_name,
                objective=objective,
                eve_time=eve_time,
                eve_corp=self.owner.corporation,
                corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
            )
        elif self.notification_type == NTYPE_MOONS_EXTRACTION_CANCELED:
            timer = None
            notifications_qs = Notification.objects\
                .filter(
                    notification_type=NTYPE_MOONS_EXTRACTION_STARTED,
                    owner=self.owner,
                    is_timer_added=True,
                    timestamp__lte=self.timestamp
                )\
                .order_by('-timestamp')

            for notification in notifications_qs:
                parsed_text_2 = notification.get_parsed_text()
                my_structure_type_id = parsed_text_2['structureTypeID']
                if my_structure_type_id == parsed_text['structureTypeID']:
                    eve_time = self._ldap_datetime_2_dt(
                        parsed_text_2['readyTime']
                    )
                    timer_query = Timer.objects\
                        .filter(
                            details=details,
                            system=system,
                            planet_moon=planet_moon,
                            structure=structure_type_name,
                            objective=objective,
                            eve_time=eve_time
                        )
                    deleted_count, _ = timer_query.delete()
                    logger.info(
                        '{}: removed {} timer related to notification'.format(
                            deleted_count,
                            self.notification_id
                        )
                    )
                    self.is_timer_added = False
                    self.save()

        return timer



from unittest.mock import Mock, patch

from bravado.exception import HTTPBadGateway
from bravado.exception import HTTPServiceUnavailable
from bravado.exception import HTTPGatewayTimeout
from bravado.exception import HTTPForbidden
    
from structures.helpers.esi_fetch import esi_fetch
from structures.helpers.esi_fetch import esi_fetch_with_localization
from structures.models.eveuniverse import EsiNameLocalization
from structures.tests.testdata import (
    esi_get_universe_categories_category_id, 
    esi_mock_client,
    esi_get_corporations_corporation_id_structures
)
from structures.utils import (
    NoSocketsTestCase, set_test_logger, make_logger_prefix
)


MODULE_PATH = __package__ + '.esi_fetch'
logger = set_test_logger(MODULE_PATH, __file__)


class TestEsiFetch(NoSocketsTestCase):
    
    def setUp(self):
        self.add_prefix = make_logger_prefix('Test')
        esi_get_corporations_corporation_id_structures.override_data = None        
    
    @patch(MODULE_PATH + '._esi_client')
    def test_can_fetch_object_from_esi(self, mock_esi_client):        
        esi_fetch(
            esi_path='Universe.get_universe_categories_category_id',
            args={'category_id': 65},
            logger_tag='dummy'
        )                
        self.assertEqual(
            len(
                mock_esi_client.return_value.Universe
                .get_universe_categories_category_id.mock_calls
            ), 
            2
        )
        args, kwargs = \
            mock_esi_client.return_value.Universe\
            .get_universe_categories_category_id.call_args
        self.assertEqual(
            kwargs, {'category_id': 65}
        )
    
    @patch(MODULE_PATH + '._esi_client')
    def test_can_fetch_object_with_client(self, mock_esi_client):
        mock_client = Mock()
        esi_fetch(
            esi_path='Universe.get_universe_categories_category_id',
            args={'category_id': 65},
            logger_tag='dummy',
            esi_client=mock_client
        )                
        self.assertEqual(
            len(mock_client.Universe.get_universe_categories_category_id.mock_calls), 
            2
        )
        args, kwargs = \
            mock_client.Universe.get_universe_categories_category_id.call_args        
        self.assertEqual(
            kwargs, {'category_id': 65}
        )    

    @patch(MODULE_PATH + '._esi_client')
    def test_can_fetch_object_with_token(self, mock_esi_client):
        mock_token = Mock()
        mock_token.access_token = 'my_access_token'        
        esi_fetch(
            esi_path='Universe.get_universe_categories_category_id',
            args={'category_id': 65},
            logger_tag='dummy',
            token=mock_token
        )                
        self.assertEqual(
            len(
                mock_esi_client.return_value.Universe
                .get_universe_categories_category_id.mock_calls
            ), 
            2
        )
        args, kwargs = \
            mock_esi_client.return_value.Universe\
            .get_universe_categories_category_id.call_args
        self.assertEqual(
            kwargs, 
            {
                'category_id': 65,
                'token': 'my_access_token'
            }
        )

    @patch(MODULE_PATH + '._esi_client')
    def test_can_fetch_object_from_esi_wo_args(self, mock_esi_client):
        mock_esi_client.side_effect = esi_mock_client
        solar_systems = esi_fetch(
            esi_path='Universe.get_universe_systems',         
            logger_tag='dummy'
        )
        expected = [30002506, 31000005, 30002537, 30000474, 30000476]
        self.assertSetEqual(set(solar_systems), set(expected))

    @patch(MODULE_PATH + '._esi_client')
    def test_raises_exception_on_invalid_esi_path(self, mock_esi_client):                        
        with self.assertRaises(ValueError):
            esi_fetch(
                'invalid', 
                {'group_id': 65}
            )
    
    @patch(MODULE_PATH + '._esi_client', spec_set=True)
    def test_raises_exception_on_wrong_esi_category(self, mock_esi_client):                        
        my_client = Mock(spec='Universe.get_universe_categories_category_id')
        mock_esi_client.return_value = my_client
        with self.assertRaises(ValueError):
            esi_fetch(
                'invalid.get_universe_groups_group_id',               
                {'group_id': 65}
            )
    
    @patch(MODULE_PATH + '._esi_client', spec_set=True)    
    def test_raises_exception_on_wrong_esi_method(self, mock_esi_client):
        my_client = Mock(spec='Universe.get_universe_categories_category_id')
        mock_esi_client.return_value = my_client
        with self.assertRaises(ValueError):
            esi_fetch(
                'Universe.invalid',                 
                {'group_id': 65}
            )
    
    @patch(MODULE_PATH + '.ESI_RETRY_SLEEP_SECS', 0)
    @patch(MODULE_PATH + '._esi_client')
    def test_can_retry_on_exceptions(self, mock_esi_client):        
        
        def my_side_effect():
            """special mock client for testing retry ability"""
            nonlocal retry_counter, max_retries
            
            if retry_counter < max_retries:                
                retry_counter += 1
                raise MyException(
                    response=Mock(), 
                    message='retry_counter=%d' % retry_counter
                )
            else:                
                return esi_get_universe_categories_category_id(category_id=65)\
                    .result()
            
        mock_esi_client.return_value.Universe\
            .get_universe_categories_category_id.return_value\
            .result.side_effect = my_side_effect

        # can retry 3 times and then proceed normally on 502s
        MyException = HTTPBadGateway
        retry_counter = 0
        max_retries = 3        
        response_object = esi_fetch(
            'Universe.get_universe_categories_category_id',
            {'category_id': 65}
        )        
        self.assertEqual(response_object['id'], 65)
        self.assertEqual(response_object['name'], 'Structure')
        self.assertEqual(retry_counter, 3)

        # will abort on the 4th retry and pass on exception if needed
        retry_counter = 0
        max_retries = 4
        with self.assertRaises(MyException):
            response_object = esi_fetch(
                'Universe.get_universe_categories_category_id',
                {'category_id': 65},
                Mock()
            )
        self.assertEqual(retry_counter, 4)

        # will retry on 503s
        MyException = HTTPServiceUnavailable
        retry_counter = 0
        max_retries = 3
        response_object = esi_fetch(
            'Universe.get_universe_categories_category_id',
            {'category_id': 65}
        )        
        self.assertEqual(response_object['id'], 65)
        self.assertEqual(response_object['name'], 'Structure')
        self.assertEqual(retry_counter, 3)

        # will retry on 504s
        MyException = HTTPGatewayTimeout
        retry_counter = 0
        max_retries = 3
        response_object = esi_fetch(
            'Universe.get_universe_categories_category_id',
            {'category_id': 65}
        )        
        self.assertEqual(response_object['id'], 65)
        self.assertEqual(response_object['name'], 'Structure')
        self.assertEqual(retry_counter, 3)

        # will not retry on other HTTP exceptions
        MyException = HTTPForbidden
        retry_counter = 0
        max_retries = 3
        with self.assertRaises(MyException):
            response_object = esi_fetch(
                'Universe.get_universe_categories_category_id',
                {'category_id': 65},
                Mock()
            )
        self.assertEqual(retry_counter, 1)
    
    def test_can_fetch_multiple_pages(self):
        mock_client = esi_mock_client()

        structures = esi_fetch(
            'Corporation.get_corporations_corporation_id_structures',
            args={'corporation_id': 2001},
            esi_client=mock_client,
            has_pages=True
        )        
        
        # has all structures
        structure_ids = {x['structure_id'] for x in structures}
        expected = {1000000000001, 1000000000002, 1000000000003}
        self.assertSetEqual(structure_ids, expected)
        
        # has services                
        service_names = list()
        for obj in structures:
            if obj['structure_id'] == 1000000000001:                        
                service_names = {x['name'] for x in obj['services']}
        expected = {'Clone Bay', 'Market Hub'}
        self.assertEqual(service_names, expected)
    
    def test_can_fetch_multiple_pages_and_languages(self):
        mock_client = esi_mock_client()

        structures_list = esi_fetch_with_localization(
            'Corporation.get_corporations_corporation_id_structures',
            args={'corporation_id': 2001},
            esi_client=mock_client,
            has_pages=True,            
            languages=EsiNameLocalization.ESI_LANGUAGES,
            logger_tag='dummy'
        )        
        
        for language, structures in structures_list.items():
            # has all structures                
            structure_ids = {x['structure_id'] for x in structures}
            expected = {1000000000001, 1000000000002, 1000000000003}
            self.assertSetEqual(structure_ids, expected)
            
            # has services in all languages
            service_names = list()
            for obj in structures:
                if obj['structure_id'] == 1000000000001:                        
                    service_names = {x['name'] for x in obj['services']}
            if language == 'en-us':
                expected = {'Clone Bay', 'Market Hub'}
            else:
                expected = {'Clone Bay_' + language, 'Market Hub_' + language}
            self.assertEqual(service_names, expected)



import logging

from time import sleep

from celery import shared_task, chain

from django.contrib.auth.models import User

from allianceauth.notifications import notify

from . import __title__
from .utils import LoggerAddTag, make_logger_prefix
from .models import Owner, Notification, Webhook, EveSovereigntyMap


logger = LoggerAddTag(logging.getLogger(__name__), __title__)


def _get_owner(owner_pk) -> Owner:
    """returns the owner or raises exception"""
    try:
        owner = Owner.objects.get(pk=owner_pk)            
    except Owner.DoesNotExist:
        raise Owner.DoesNotExist(
            "Requested owner with pk {} does not exist".format(owner_pk)
        )    
    return owner


def _get_user(user_pk) -> User:
    """returns the user or None. Logs if user is requested but can't be found."""
    user = None
    if user_pk:
        try:
            user = User.objects.get(pk=user_pk)
        except User.DoesNotExist:
            logger.warning(
                'Ignoring non-existing user with pk {}'.format(user_pk)
            )            
    return user


@shared_task
def update_structures_for_owner(owner_pk, user_pk=None):
    """fetches all structures for owner from ESI"""
    _get_owner(owner_pk).update_structures_esi(_get_user(user_pk))


@shared_task
def update_structures():
    """fetches all structures for all active owner from ESI"""    
    for owner in Owner.objects.all():
        if owner.is_active:
            update_structures_for_owner.delay(owner.pk)


@shared_task
def update_sov_map():
    """updates the sovereignty map"""
    EveSovereigntyMap.objects.update_from_esi()


@shared_task
def update_all_structures():
    """main task for starting regular update of all structures 
    and related data from ESI
    """
    chain(update_sov_map.si(), update_structures.si()).delay()


@shared_task
def fetch_notifications_for_owner(owner_pk, user_pk=None):
    """fetches all notification for owner from ESI and proceses them"""
    _get_owner(owner_pk).fetch_notifications_esi(_get_user(user_pk))
    

@shared_task
def fetch_all_notifications():
    """fetch notifications for all owners"""
    for owner in Owner.objects.all():
        if owner.is_active:
            fetch_notifications_for_owner.apply_async(
                kwargs={'owner_pk': owner.pk}, priority=2
            )
            

@shared_task
def send_new_notifications_for_owner(owner_pk, rate_limited=True, user_pk=None):
    """forwards new notification for this owner to configured webhooks"""
    _get_owner(owner_pk).send_new_notifications(rate_limited, _get_user(user_pk))


@shared_task
def send_all_new_notifications(rate_limited=True):
    """sends all unsent notifications to active webhooks and adds timers"""
    send_tasks = list()
    for owner in Owner.objects.all():
        if owner.is_active:
            send_tasks.append(send_new_notifications_for_owner.si(
                owner_pk=owner.pk, rate_limited=rate_limited
            ))

    chain(send_tasks).apply_async(priority=2)


@shared_task
def send_notifications(notification_pks: list, rate_limited=True):
    """send notifications defined by list of pks"""    
    notifications = Notification.objects.filter(pk__in=notification_pks)
    if notifications:
        logger.info('Trying to send {} notifications to webhooks...'.format(
            len(notification_pks)
        ))        
        for n in notifications:
            for webhook in n.owner.webhooks.all():
                if (str(n.notification_type) in webhook.notification_types
                    and not n.filter_for_npc_attacks()
                    and not n.filter_for_alliance_level()
                ):
                    n.send_to_webhook(webhook)
            if rate_limited:
                sleep(1)


@shared_task
def send_test_notifications_to_webhook(webhook_pk, user_pk=None):
    """sends test notification to given webhook"""

    add_prefix = make_logger_prefix('test notification')
    try:
        webhook = Webhook.objects.get(pk=webhook_pk)
        add_prefix = make_logger_prefix(webhook)
        send_report = webhook.send_test_notification()
        error_code = None
    except Exception as ex:
        logger.exception('Failed to send test notification')
        send_report = None
        error_code = str(ex)

    success = (error_code is None)
    if user_pk:
        try:
            message = 'Test notification to webhook "{}" {}.\n'.format(
                webhook,
                'completed successfully' if success else 'has failed'
            )
            if success:
                message += 'send report:\n{}'.format(send_report)
            else:
                message += 'Error code: {}'.format(error_code)

            notify(
                user=User.objects.get(pk=user_pk),
                title='{}: Test notification to "{}": {}'.format(
                    __title__,
                    webhook,
                    'OK' if success else 'FAILED'
                ),
                message=message,
                level='success' if success else 'danger'
            )
        except Exception as ex:
            logger.exception(add_prefix(
                'An unexpected error ocurred while trying to '
                + 'report to user: {}'. format(ex)
            ))



# flake8: noqa

from .owners import Owner
from .eveuniverse import (        
    EsiNameLocalization,
    EveCategory, 
    EveGroup, 
    EveType, 
    EveRegion, 
    EveConstellation, 
    EveSolarSystem, 
    EvePlanet, 
    EveMoon,
    EveSovereigntyMap  
)
from .structures import (
    Structure, StructureService, StructureTag
)  
from .notifications import (
    Webhook, Notification, EveEntity, get_default_notification_types
)


from django.conf.urls import url
from . import views


app_name = 'structures'

urlpatterns = [
    url(r'^$', views.index, name='index'),
    url(r'^list/$', views.structure_list, name='structure_list'),
    url(
        r'^list_data/$', views.structure_list_data, name='structure_list_data'
    ),
    url(
        r'^add_structure_owner/$',
        views.add_structure_owner,
        name='add_structure_owner'
    ),
    url(r'^service_status/$', views.service_status, name='service_status'),
]



"""Structure related models"""

import re
import logging

from django.db import models
from django.core.validators import MinValueValidator
from django.core.validators import MaxValueValidator
from django.utils.html import escape
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop

from .. import __title__
from .eveuniverse import EsiNameLocalization
from .eveuniverse import EveSolarSystem
from ..managers import StructureManager
from ..managers import StructureTagManager
from ..utils import LoggerAddTag, create_bs_label_html


logger = LoggerAddTag(logging.getLogger(__name__), __title__)


class StructureTag(models.Model):
    """tag for organizing structures"""

    # special tags
    NAME_SOV_TAG = gettext_noop('sov')
    NAME_HIGHSEC_TAG = gettext_noop('highsec')
    NAME_LOWSEC_TAG = gettext_noop('lowsec')
    NAME_NULLSEC_TAG = gettext_noop('nullsec')
    NAME_W_SPACE_TAG = gettext_noop('w_space')
    
    # styles
    STYLE_GREY = 'default'
    STYLE_DARK_BLUE = 'primary'
    STYLE_GREEN = 'success'
    STYLE_LIGHT_BLUE = 'info'
    STYLE_ORANGE = 'warning'
    STYLE_RED = 'danger'
    STYLE_CHOICES = [
        (STYLE_GREY, 'grey'),
        (STYLE_DARK_BLUE, 'dark blue'),
        (STYLE_GREEN, 'green'),
        (STYLE_LIGHT_BLUE, 'light blue'),
        (STYLE_ORANGE, 'orange'),
        (STYLE_RED, 'red'),
    ]
    
    SPACE_TYPE_MAP = {
        EveSolarSystem.TYPE_HIGHSEC: {
            'name': NAME_HIGHSEC_TAG,
            'style': STYLE_GREEN
        },
        EveSolarSystem.TYPE_LOWSEC: {
            'name': NAME_LOWSEC_TAG,
            'style': STYLE_ORANGE
        },
        EveSolarSystem.TYPE_NULLSEC: {
            'name': NAME_NULLSEC_TAG,
            'style': STYLE_RED
        },
        EveSolarSystem.TYPE_W_SPACE: {
            'name': NAME_W_SPACE_TAG,
            'style': STYLE_LIGHT_BLUE
        },        
    }

    name = models.CharField(
        max_length=255,
        unique=True,
        help_text='name of the tag - must be unique'
    )
    description = models.TextField(
        null=True,
        default=None,
        blank=True,
        help_text='description for this tag'
    )
    style = models.CharField(
        max_length=16,
        choices=STYLE_CHOICES,
        default='default',
        blank=True,
        help_text='color style of tag'
    )
    order = models.PositiveIntegerField(
        default=100,
        blank=True,
        validators=[MinValueValidator(100)],
        help_text=(
            'number defining the order tags are shown. '
            'custom tags can not have an order below 100'
        )
    )    
    is_default = models.BooleanField(
        default=False,
        help_text=(
            'if true this custom tag will automatically be added '
            'to new structures'
        )
    )
    is_user_managed = models.BooleanField(
        default=True,
        help_text=(
            'if False this tag is created and managed by the system '
            'and can not be modified by users'
        )
    )

    objects = StructureTagManager()

    def __str__(self) -> str:
        return self.name

    def __repr__(self):
        return '{}(name=\'{}\')'.format(
            self.__class__.__name__,            
            self.name
        )

    class Meta:
        ordering = ordering = ['order', 'name']

    @property
    def html(self) -> str:
        if self.is_user_managed:
            name = escape(self.name)
        else:
            name = _(self.name)
        return create_bs_label_html(name, self.style)

    @classmethod
    def sorted(cls, tags: list, reverse: bool = False) -> list:
        """returns a sorted copy of the provided list of tags"""
        return sorted(tags, key=lambda x: x.name.lower(), reverse=reverse)


class Structure(models.Model):
    """structure of a corporation"""

    STATE_NA = 0
    STATE_ANCHOR_VULNERABLE = 1
    STATE_ANCHORING = 2
    STATE_ARMOR_REINFORCE = 3
    STATE_ARMOR_VULNERABLE = 4
    STATE_DEPLOY_VULNERABLE = 5
    STATE_FITTING_INVULNERABLE = 6
    STATE_HULL_REINFORCE = 7
    STATE_HULL_VULNERABLE = 8
    STATE_ONLINE_DEPRECATED = 9
    STATE_ONLINING_VULNERABLE = 10
    STATE_SHIELD_VULNERABLE = 11
    STATE_UNANCHORED = 12
    STATE_UNKNOWN = 13

    STATE_POS_OFFLINE = 21
    STATE_POS_ONLINE = 22
    STATE_POS_ONLINING = 23
    STATE_POS_REINFORCED = 24
    STATE_POS_UNANCHORING = 25

    STATE_CHOICES = [
        # states Upwell structures
        (STATE_ANCHOR_VULNERABLE, _('anchor vulnerable')),
        (STATE_ANCHORING, _('anchoring')),
        (STATE_ARMOR_REINFORCE, _('armor reinforce')),
        (STATE_ARMOR_VULNERABLE, _('armor vulnerable')),
        (STATE_DEPLOY_VULNERABLE, _('deploy vulnerable')),
        (STATE_FITTING_INVULNERABLE, _('fitting invulnerable')),
        (STATE_HULL_REINFORCE, _('hull reinforce')),
        (STATE_HULL_VULNERABLE, _('hull vulnerable')),
        (STATE_ONLINE_DEPRECATED, _('online deprecated')),
        (STATE_ONLINING_VULNERABLE, _('onlining vulnerable')),
        (STATE_SHIELD_VULNERABLE, _('shield vulnerable')),
        (STATE_UNANCHORED, _('unanchored')),

        # starbases
        (STATE_POS_OFFLINE, _('offline')),
        (STATE_POS_ONLINE, _('online')),
        (STATE_POS_ONLINING, _('onlining')),
        (STATE_POS_REINFORCED, _('reinforced')),
        (STATE_POS_UNANCHORING, _('unanchoring ')),

        # other
        (STATE_NA, _('N/A')),
        (STATE_UNKNOWN, _('unknown')),
    ]
    _STATES_ESI_MAP = {        
        'anchor_vulnerable': STATE_ANCHOR_VULNERABLE,
        'anchoring': STATE_ANCHORING,
        'armor_reinforce': STATE_ARMOR_REINFORCE,
        'armor_vulnerable': STATE_ARMOR_VULNERABLE,
        'deploy_vulnerable': STATE_DEPLOY_VULNERABLE,
        'fitting_invulnerable': STATE_FITTING_INVULNERABLE,
        'hull_reinforce': STATE_HULL_REINFORCE,
        'hull_vulnerable': STATE_HULL_VULNERABLE,
        'online_deprecated': STATE_ONLINE_DEPRECATED,
        'onlining_vulnerable': STATE_ONLINING_VULNERABLE,
        'shield_vulnerable': STATE_SHIELD_VULNERABLE,
        'unanchored': STATE_UNANCHORED,
        'offline': STATE_POS_OFFLINE,
        'online': STATE_POS_ONLINE,
        'onlining': STATE_POS_ONLINING,
        'reinforced': STATE_POS_REINFORCED,
        'unanchoring ': STATE_POS_UNANCHORING,
    }

    id = models.BigIntegerField(
        primary_key=True,
        help_text='The Item ID of the structure'
    )
    owner = models.ForeignKey(
        'Owner',
        on_delete=models.CASCADE,
        help_text='Corporation that owns the structure'
    )
    eve_type = models.ForeignKey(
        'EveType',
        on_delete=models.CASCADE,
        help_text='type of the structure'
    )
    name = models.CharField(
        max_length=255,
        help_text='The full name of the structure'
    )
    eve_solar_system = models.ForeignKey(
        'EveSolarSystem',
        on_delete=models.CASCADE
    )
    eve_planet = models.ForeignKey(
        'EvePlanet',
        on_delete=models.SET_DEFAULT,
        null=True,
        default=None,
        blank=True,
        help_text='Planet next to this structure - if any'
    )
    eve_moon = models.ForeignKey(
        'EveMoon',
        on_delete=models.SET_DEFAULT,
        null=True,
        default=None,
        blank=True,
        help_text='Moon next to this structure - if any'
    )
    position_x = models.FloatField(
        null=True,
        default=None,
        blank=True,
        help_text='x position in the solar system'
    )
    position_y = models.FloatField(
        null=True,
        default=None,
        blank=True,
        help_text='y position in the solar system'
    )
    position_z = models.FloatField(
        null=True,
        default=None,
        blank=True,
        help_text='z position in the solar system'
    )
    fuel_expires = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        help_text='Date on which the structure will run out of fuel'
    )
    next_reinforce_hour = models.PositiveIntegerField(
        null=True,
        default=None,
        blank=True,
        validators=[MaxValueValidator(23)],
        help_text=(
            'The requested change to reinforce_hour that will take '
            'effect at the time shown by next_reinforce_apply'
        )
    )
    next_reinforce_weekday = models.PositiveIntegerField(
        null=True,
        default=None,
        blank=True,
        validators=[MaxValueValidator(6)],
        help_text=(
            'The date and time when the structure’s newly requested '
            'reinforcement times (e.g. next_reinforce_hour and '
            'next_reinforce_day) will take effect'
        )
    )
    next_reinforce_apply = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        help_text=(
            'The requested change to reinforce_weekday that will take '
            'effect at the time shown by next_reinforce_apply'
        )
    )
    reinforce_hour = models.PositiveIntegerField(
        validators=[MaxValueValidator(23)],
        null=True,
        default=None,
        blank=True,
        help_text=(
            'The hour of day that determines the four hour window '
            'when the structure will randomly exit its reinforcement periods '
            'and become vulnerable to attack against its armor and/or hull. '
            'The structure will become vulnerable at a random time that '
            'is +/- 2 hours centered on the value of this property'
        )
    )
    reinforce_weekday = models.PositiveIntegerField(
        null=True,
        default=None,
        blank=True,
        validators=[MaxValueValidator(6)],
        help_text='(no longer used)'
    )
    state = models.IntegerField(
        choices=STATE_CHOICES,
        default=STATE_UNKNOWN,
        blank=True,
        help_text='Current state of the structure'
    )
    state_timer_start = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        help_text='Date at which the structure will move to it’s next state'
    )
    state_timer_end = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        help_text='Date at which the structure entered it’s current state'
    )
    unanchors_at = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        help_text='Date at which the structure will unanchor'
    )
    last_updated = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        help_text='date this structure was last updated from the EVE server'
    )
    tags = models.ManyToManyField(
        StructureTag,
        default=None,
        blank=True,
        help_text='list of tags for this structure'
    )

    objects = StructureManager()

    @property
    def is_low_power(self):
        return False if not self.eve_type.is_upwell_structure \
            else not self.fuel_expires

    @property
    def is_reinforced(self):
        return self.state in [
            self.STATE_ARMOR_REINFORCE,
            self.STATE_HULL_REINFORCE,
            self.STATE_ANCHOR_VULNERABLE,
            self.STATE_HULL_VULNERABLE
        ]

    @property
    def owner_has_sov(self):
        return self.eve_solar_system.corporation_has_sov(self.owner.corporation)

    def __str__(self):
        return '{} - {}'.format(self.eve_solar_system, self.name)

    def __repr__(self):
        return '{}(id={}, name=\'{}\')'.format(
            self.__class__.__name__,
            self.id,
            self.name
        )

    @classmethod
    def get_matching_state_for_esi_state(cls, esi_state_name) -> int:
        """returns matching state for esi state name of Upwell structures"""
        return (
            cls._STATES_ESI_MAP[esi_state_name]
            if esi_state_name in cls._STATES_ESI_MAP
            else cls.STATE_UNKNOWN
        )

    @classmethod
    def extract_name_from_esi_respose(cls, esi_name):
        """extracts the structure's name from the name in an ESI response"""
        matches = re.search(r'^\S+ - (.+)', esi_name)
        return matches.group(1) if matches else esi_name

    def save(self, *args, **kwargs):
        """make sure related objects are saved whenever structure is saved"""
        super().save(*args, **kwargs)
        self.update_generated_tags()
        
    def update_generated_tags(self, recreate_tags=False):
        """updates all generated tags for this structure
        
        recreate_tags: when set true all tags will be re-created,
        otherwise just re-added if they are missing
        """        
        method_name = 'update_or_create_for_space_type' \
            if recreate_tags else 'get_or_create_for_space_type'

        space_type_tag, _ = getattr(StructureTag.objects, method_name)(
            self.eve_solar_system
        )
        
        self.tags.add(space_type_tag)
        if self.owner_has_sov:
            method_name = 'update_or_create_for_sov' \
                if recreate_tags else 'get_or_create_for_sov'

            sov_tag, _ = getattr(StructureTag.objects, method_name)()
            self.tags.add(sov_tag)
        
        
class StructureService(EsiNameLocalization, models.Model):
    """service of a structure"""

    STATE_OFFLINE = 1
    STATE_ONLINE = 2

    STATE_CHOICES = [
        (STATE_OFFLINE, _('offline')),
        (STATE_ONLINE, _('online')),
    ]

    _STATES_ESI_MAP = {
        'offline': STATE_OFFLINE,
        'online': STATE_ONLINE,
    }

    structure = models.ForeignKey(
        Structure,
        on_delete=models.CASCADE,
        help_text='Structure this service is installed to'
    )
    name = models.CharField(
        max_length=100,
        help_text='Name of the service'
    )
    state = models.IntegerField(
        choices=STATE_CHOICES,
        help_text='Current state of this service'
    )

    class Meta:
        unique_together = (('structure', 'name'),)

    def __str__(self):
        return '{} - {}'.format(str(self.structure), self.name)

    def __repr__(self):
        return '{}(structure_id={}, name=\'{}\')'.format(
            self.__class__.__name__,
            self.structure.id,
            self.name
        )

    @classmethod
    def get_matching_state_for_esi_state(cls, esi_state_name) -> int:
        """returns matching state for given state name"""
        return (
            cls._STATES_ESI_MAP[esi_state_name]
            if esi_state_name in cls._STATES_ESI_MAP
            else cls.STATE_OFFLINE
        )



# Generated by Django 2.2.11 on 2020-04-13 20:02

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structures', '0009_localizations'),
    ]

    operations = [
        migrations.CreateModel(
            name='EveSovereigntyMap',
            fields=[
                ('solar_system_id', models.PositiveIntegerField(primary_key=True, serialize=False)),
                ('alliance_id', models.PositiveIntegerField(blank=True, db_index=True, help_text='alliance who holds sov for this system', null=True)),
                ('corporation_id', models.PositiveIntegerField(blank=True, db_index=True, help_text='corporation who holds sov for this system', null=True)),
                ('faction_id', models.PositiveIntegerField(blank=True, db_index=True, help_text='faction who holds sov for this system', null=True)),
                ('last_updated', models.DateTimeField(blank=True, db_index=True, default=None, help_text='When this object was last updated from ESI', null=True)),
            ],
        ),
        migrations.AlterModelOptions(
            name='structuretag',
            options={'ordering': ['order', 'name']},
        ),
        migrations.AddField(
            model_name='structuretag',
            name='is_user_managed',
            field=models.BooleanField(default=True, help_text='if False this tag is created and managed by the system and can not be modified by users'),
        ),
        migrations.AddField(
            model_name='structuretag',
            name='order',
            field=models.PositiveIntegerField(blank=True, default=100, help_text='number defining the order tags are shown. custom tags can not have an order below 100', validators=[django.core.validators.MinValueValidator(100)]),
        ),
        migrations.AlterField(
            model_name='structuretag',
            name='is_default',
            field=models.BooleanField(default=False, help_text='if true this custom tag will automatically be added to new structures'),
        ),
        migrations.AlterField(
            model_name='structuretag',
            name='style',
            field=models.CharField(blank=True, choices=[('default', 'grey'), ('primary', 'dark blue'), ('success', 'green'), ('info', 'light blue'), ('warning', 'orange'), ('danger', 'red')], default='default', help_text='color style of tag', max_length=16),
        ),
    ]



# Generated by Django 2.2.11 on 2020-03-31 12:26

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structures', '0008_auto_20200209_1638'),
    ]

    operations = [
        migrations.AddField(
            model_name='evecategory',
            name='last_updated',
            field=models.DateTimeField(blank=True, db_index=True, default=None, help_text='When this object was last updated from ESI', null=True),
        ),
        migrations.AddField(
            model_name='evecategory',
            name='name_de',
            field=models.CharField(blank=True, help_text='Eve Online name localized for German', max_length=100),
        ),
        migrations.AddField(
            model_name='evecategory',
            name='name_ko',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Korean', max_length=100),
        ),
        migrations.AddField(
            model_name='evecategory',
            name='name_ru',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Russian', max_length=100),
        ),
        migrations.AddField(
            model_name='evecategory',
            name='name_zh',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Chinese', max_length=100),
        ),
        migrations.AddField(
            model_name='eveconstellation',
            name='last_updated',
            field=models.DateTimeField(blank=True, db_index=True, default=None, help_text='When this object was last updated from ESI', null=True),
        ),
        migrations.AddField(
            model_name='eveconstellation',
            name='name_de',
            field=models.CharField(blank=True, help_text='Eve Online name localized for German', max_length=100),
        ),
        migrations.AddField(
            model_name='eveconstellation',
            name='name_ko',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Korean', max_length=100),
        ),
        migrations.AddField(
            model_name='eveconstellation',
            name='name_ru',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Russian', max_length=100),
        ),
        migrations.AddField(
            model_name='eveconstellation',
            name='name_zh',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Chinese', max_length=100),
        ),
        migrations.AddField(
            model_name='evegroup',
            name='last_updated',
            field=models.DateTimeField(blank=True, db_index=True, default=None, help_text='When this object was last updated from ESI', null=True),
        ),
        migrations.AddField(
            model_name='evegroup',
            name='name_de',
            field=models.CharField(blank=True, help_text='Eve Online name localized for German', max_length=100),
        ),
        migrations.AddField(
            model_name='evegroup',
            name='name_ko',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Korean', max_length=100),
        ),
        migrations.AddField(
            model_name='evegroup',
            name='name_ru',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Russian', max_length=100),
        ),
        migrations.AddField(
            model_name='evegroup',
            name='name_zh',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Chinese', max_length=100),
        ),
        migrations.AddField(
            model_name='evemoon',
            name='last_updated',
            field=models.DateTimeField(blank=True, db_index=True, default=None, help_text='When this object was last updated from ESI', null=True),
        ),
        migrations.AddField(
            model_name='evemoon',
            name='name_de',
            field=models.CharField(blank=True, help_text='Eve Online name localized for German', max_length=100),
        ),
        migrations.AddField(
            model_name='evemoon',
            name='name_ko',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Korean', max_length=100),
        ),
        migrations.AddField(
            model_name='evemoon',
            name='name_ru',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Russian', max_length=100),
        ),
        migrations.AddField(
            model_name='evemoon',
            name='name_zh',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Chinese', max_length=100),
        ),
        migrations.AddField(
            model_name='eveplanet',
            name='last_updated',
            field=models.DateTimeField(blank=True, db_index=True, default=None, help_text='When this object was last updated from ESI', null=True),
        ),
        migrations.AddField(
            model_name='eveplanet',
            name='name_de',
            field=models.CharField(blank=True, help_text='Eve Online name localized for German', max_length=100),
        ),
        migrations.AddField(
            model_name='eveplanet',
            name='name_ko',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Korean', max_length=100),
        ),
        migrations.AddField(
            model_name='eveplanet',
            name='name_ru',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Russian', max_length=100),
        ),
        migrations.AddField(
            model_name='eveplanet',
            name='name_zh',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Chinese', max_length=100),
        ),
        migrations.AddField(
            model_name='everegion',
            name='last_updated',
            field=models.DateTimeField(blank=True, db_index=True, default=None, help_text='When this object was last updated from ESI', null=True),
        ),
        migrations.AddField(
            model_name='everegion',
            name='name_de',
            field=models.CharField(blank=True, help_text='Eve Online name localized for German', max_length=100),
        ),
        migrations.AddField(
            model_name='everegion',
            name='name_ko',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Korean', max_length=100),
        ),
        migrations.AddField(
            model_name='everegion',
            name='name_ru',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Russian', max_length=100),
        ),
        migrations.AddField(
            model_name='everegion',
            name='name_zh',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Chinese', max_length=100),
        ),
        migrations.AddField(
            model_name='evesolarsystem',
            name='last_updated',
            field=models.DateTimeField(blank=True, db_index=True, default=None, help_text='When this object was last updated from ESI', null=True),
        ),
        migrations.AddField(
            model_name='evesolarsystem',
            name='name_de',
            field=models.CharField(blank=True, help_text='Eve Online name localized for German', max_length=100),
        ),
        migrations.AddField(
            model_name='evesolarsystem',
            name='name_ko',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Korean', max_length=100),
        ),
        migrations.AddField(
            model_name='evesolarsystem',
            name='name_ru',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Russian', max_length=100),
        ),
        migrations.AddField(
            model_name='evesolarsystem',
            name='name_zh',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Chinese', max_length=100),
        ),
        migrations.AddField(
            model_name='evetype',
            name='last_updated',
            field=models.DateTimeField(blank=True, db_index=True, default=None, help_text='When this object was last updated from ESI', null=True),
        ),
        migrations.AddField(
            model_name='evetype',
            name='name_de',
            field=models.CharField(blank=True, help_text='Eve Online name localized for German', max_length=100),
        ),
        migrations.AddField(
            model_name='evetype',
            name='name_ko',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Korean', max_length=100),
        ),
        migrations.AddField(
            model_name='evetype',
            name='name_ru',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Russian', max_length=100),
        ),
        migrations.AddField(
            model_name='evetype',
            name='name_zh',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Chinese', max_length=100),
        ),
        migrations.AddField(
            model_name='structureservice',
            name='name_de',
            field=models.CharField(blank=True, help_text='Eve Online name localized for German', max_length=100),
        ),
        migrations.AddField(
            model_name='structureservice',
            name='name_ko',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Korean', max_length=100),
        ),
        migrations.AddField(
            model_name='structureservice',
            name='name_ru',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Russian', max_length=100),
        ),
        migrations.AddField(
            model_name='structureservice',
            name='name_zh',
            field=models.CharField(blank=True, help_text='Eve Online name localized for Chinese', max_length=100),
        ),
        migrations.AddField(
            model_name='webhook',
            name='language_code',
            field=models.CharField(blank=True, choices=[('en', 'English'), ('de', 'German'), ('es', 'Spanish'), ('zh-hans', 'Chinese Simplified'), ('ru', 'Russian'), ('ko', 'Korean')], default=None, help_text='language of notifications send to this webhook', max_length=8, null=True, verbose_name='language'),
        ),
        migrations.AlterField(
            model_name='evecategory',
            name='id',
            field=models.PositiveIntegerField(help_text='Eve Online ID', primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='evecategory',
            name='name',
            field=models.CharField(help_text='Eve Online name', max_length=100),
        ),
        migrations.AlterField(
            model_name='eveconstellation',
            name='id',
            field=models.PositiveIntegerField(help_text='Eve Online ID', primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='eveconstellation',
            name='name',
            field=models.CharField(help_text='Eve Online name', max_length=100),
        ),
        migrations.AlterField(
            model_name='eveentity',
            name='id',
            field=models.PositiveIntegerField(help_text='Eve Online ID', primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='evegroup',
            name='id',
            field=models.PositiveIntegerField(help_text='Eve Online ID', primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='evegroup',
            name='name',
            field=models.CharField(help_text='Eve Online name', max_length=100),
        ),
        migrations.AlterField(
            model_name='evemoon',
            name='id',
            field=models.PositiveIntegerField(help_text='Eve Online ID', primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='evemoon',
            name='name',
            field=models.CharField(help_text='Eve Online name', max_length=100),
        ),
        migrations.AlterField(
            model_name='evemoon',
            name='position_x',
            field=models.FloatField(blank=True, default=None, help_text='x position in the solar system', null=True),
        ),
        migrations.AlterField(
            model_name='evemoon',
            name='position_y',
            field=models.FloatField(blank=True, default=None, help_text='y position in the solar system', null=True),
        ),
        migrations.AlterField(
            model_name='evemoon',
            name='position_z',
            field=models.FloatField(blank=True, default=None, help_text='z position in the solar system', null=True),
        ),
        migrations.AlterField(
            model_name='eveplanet',
            name='id',
            field=models.PositiveIntegerField(help_text='Eve Online ID', primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='eveplanet',
            name='name',
            field=models.CharField(help_text='Eve Online name', max_length=100),
        ),
        migrations.AlterField(
            model_name='eveplanet',
            name='position_x',
            field=models.FloatField(blank=True, default=None, help_text='x position in the solar system', null=True),
        ),
        migrations.AlterField(
            model_name='eveplanet',
            name='position_y',
            field=models.FloatField(blank=True, default=None, help_text='y position in the solar system', null=True),
        ),
        migrations.AlterField(
            model_name='eveplanet',
            name='position_z',
            field=models.FloatField(blank=True, default=None, help_text='z position in the solar system', null=True),
        ),
        migrations.AlterField(
            model_name='everegion',
            name='id',
            field=models.PositiveIntegerField(help_text='Eve Online ID', primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='everegion',
            name='name',
            field=models.CharField(help_text='Eve Online name', max_length=100),
        ),
        migrations.AlterField(
            model_name='evesolarsystem',
            name='id',
            field=models.PositiveIntegerField(help_text='Eve Online ID', primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='evesolarsystem',
            name='name',
            field=models.CharField(help_text='Eve Online name', max_length=100),
        ),
        migrations.AlterField(
            model_name='evetype',
            name='id',
            field=models.PositiveIntegerField(help_text='Eve Online ID', primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='evetype',
            name='name',
            field=models.CharField(help_text='Eve Online name', max_length=100),
        ),
        migrations.AlterField(
            model_name='notification',
            name='is_read',
            field=models.BooleanField(blank=True, default=None, help_text='True when this notification has read in the eve client', null=True),
        ),
        migrations.AlterField(
            model_name='notification',
            name='text',
            field=models.TextField(blank=True, default=None, help_text='Notification details in YAML', null=True),
        ),
        migrations.AlterField(
            model_name='structure',
            name='next_reinforce_hour',
            field=models.PositiveIntegerField(blank=True, default=None, help_text='The requested change to reinforce_hour that will take effect at the time shown by next_reinforce_apply', null=True, validators=[django.core.validators.MaxValueValidator(23)]),
        ),
        migrations.AlterField(
            model_name='structure',
            name='next_reinforce_weekday',
            field=models.PositiveIntegerField(blank=True, default=None, help_text='The date and time when the structure’s newly requested reinforcement times (e.g. next_reinforce_hour and next_reinforce_day) will take effect', null=True, validators=[django.core.validators.MaxValueValidator(6)]),
        ),
        migrations.AlterField(
            model_name='structure',
            name='position_x',
            field=models.FloatField(blank=True, default=None, help_text='x position in the solar system', null=True),
        ),
        migrations.AlterField(
            model_name='structure',
            name='position_y',
            field=models.FloatField(blank=True, default=None, help_text='y position in the solar system', null=True),
        ),
        migrations.AlterField(
            model_name='structure',
            name='position_z',
            field=models.FloatField(blank=True, default=None, help_text='z position in the solar system', null=True),
        ),
        migrations.AlterField(
            model_name='structure',
            name='reinforce_hour',
            field=models.PositiveIntegerField(blank=True, default=None, help_text='The hour of day that determines the four hour window when the structure will randomly exit its reinforcement periods and become vulnerable to attack against its armor and/or hull. The structure will become vulnerable at a random time that is +/- 2 hours centered on the value of this property', null=True, validators=[django.core.validators.MaxValueValidator(23)]),
        ),
        migrations.AlterField(
            model_name='structure',
            name='reinforce_weekday',
            field=models.PositiveIntegerField(blank=True, default=None, help_text='(no longer used)', null=True, validators=[django.core.validators.MaxValueValidator(6)]),
        ),
        migrations.AlterField(
            model_name='structure',
            name='state',
            field=models.IntegerField(blank=True, choices=[(1, 'anchor vulnerable'), (2, 'anchoring'), (3, 'armor reinforce'), (4, 'armor vulnerable'), (5, 'deploy vulnerable'), (6, 'fitting invulnerable'), (7, 'hull reinforce'), (8, 'hull vulnerable'), (9, 'online deprecated'), (10, 'onlining vulnerable'), (11, 'shield vulnerable'), (12, 'unanchored'), (21, 'offline'), (22, 'online'), (23, 'onlining'), (24, 'reinforced'), (25, 'unanchoring '), (0, 'N/A'), (13, 'unknown')], default=13, help_text='Current state of the structure'),
        ),
        migrations.AlterField(
            model_name='structureservice',
            name='name',
            field=models.CharField(help_text='Name of the service', max_length=100),
        ),
        migrations.AlterField(
            model_name='structuretag',
            name='is_default',
            field=models.BooleanField(default=False, help_text='if true this tag will automatically be added to new structures'),
        ),
        migrations.AlterField(
            model_name='webhook',
            name='is_default',
            field=models.BooleanField(default=False, help_text='whether owners have this webhook automatically pre-set when created'),
        ),
    ]
